# Copyright 2014 Tom Hayward <tom@tomh.us>
# Copyright 2014 Jens Jensen <af5mi@yahoo.com>
# Copyright 2014 James Lee N1DDK <jml@jmlzone.com>
# Copyright 2016 Nathan Crapo <nathan_crapo@yahoo.com>  (TH-7800 only)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import bitwise, chirp_common, directory, errors, util, memmap
import struct
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueBoolean, \
    RadioSettingValueString, RadioSettings, \
    RadioSettingValueMap, zero_indexed_seq_map
from chirp.chirp_common import format_freq
import logging
from datetime import date

LOG = logging.getLogger(__name__)

TH7800_MEM_FORMAT = """
struct mem {
  lbcd rx_freq[4];
  lbcd tx_freq[4];
  lbcd ctcss[2];
  lbcd dtcs[2];
  u8 power:2,
     clk_sft:1,
     unknown0a:2,
     display:1,     // freq=0, name=1
     scan:2;
  u8 fmdev:2,       // wide=00, mid=01, narrow=10
     scramb:1,
     compand:1,
     emphasis:1,
     unknown1a:2,
     sqlmode:1;     // carrier, tone
  u8 rptmod:2,      // off, -, +
     reverse:1,
     talkaround:1,
     step:4;
  u8 dtcs_pol:2,
     unknown3:4,
     tmode:2;
  lbcd offset[4];
  u8 hsdtype:2,     // off, 2-tone, 5-tone, dtmf
     unknown5a:1,
     am:1,
     unknown5b:4;
  u8 unknown6[3];
  char name[6];
  u8 empty[2];
};

#seekto 0x%04X;
struct mem memory[800];

#seekto 0x%04X;
struct {
  struct mem lower;
  struct mem upper;
} scanlimits[5];

#seekto 0x%04X;
struct {
    u8  unk0xdc20:5,
        left_sql:3;
    u8  apo;
    u8  unk0xdc22:5,
        backlight:3;
    u8  unk0xdc23;
    u8  beep:1,
        keylock:1,
        pttlock:2,
        unk0xdc24_32:2,
        hyper_chan:1,
        right_func_key:1;
    u8  tbst_freq:2,
        unk0xdc25_4:2,
        mute_mode:2,
        unk0xdc25_10:2;
    u8  ars:1,
        unk0xdc26_54:3,
        auto_am:1,
        unk0xdc26_210:3;
    u8  unk0xdc27_76543:5,
        scan_mode:1,
        unk0xdc27_1:1,
        scan_resume:1;
    u16 scramb_freq;
    u16 scramb_freq1;
    u8  unk0xdc2c;
    u8  unk0xdc2d;
    u8  unk0xdc2e:5,
        right_sql:3;
    u8  unk0xdc2f:8;
    u8  tot;
    u8  unk0xdc30;
    u8  unk0xdc31;
    u8  unk0xdc32;
    u8  unk0xdc34;
    u8  unk0xdc35;
    u8  unk0xdc36;
    u8  unk0xdc37;
    u8  p1;
    u8  p2;
    u8  p3;
    u8  p4;
} settings;

#seekto 0x%04X;
u8  chan_active[128];
u8  scan_enable[128];
u8  priority[128];

#seekto 0x%04X;
struct {
    char sn[8];
    char model[8];
    char code[16];
    u8 empty[8];
    lbcd prog_yr[2];
    lbcd prog_mon;
    lbcd prog_day;
    u8 empty_10f2c[4];
} info;

struct {
  lbcd lorx[4];
  lbcd hirx[4];
  lbcd lotx[4];
  lbcd hitx[4];
} bandlimits[9];

"""


BLANK_MEMORY = "\xFF" * 8 + "\x00\x10\x23\x00\xC0\x08\x06\x00" \
               "\x00\x00\x76\x00\x00\x00" + "\xFF" * 10
DTCS_POLARITY = ["NN", "RN", "NR", "RR"]
SCAN_MODES = ["", "S", "P"]
MODES = ["WFM", "FM", "NFM"]
TMODES = ["", "Tone", "TSQL", "DTCS"]
POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=5.00),
                chirp_common.PowerLevel("Mid2", watts=10.00),
                chirp_common.PowerLevel("Mid1", watts=20.00),
                chirp_common.PowerLevel("High", watts=50.00)]
BUSY_LOCK = ["off", "Carrier", "2 tone"]
MICKEYFUNC = ["None", "SCAN", "SQL.OFF", "TCALL", "PPTR", "PRI", "LOW", "TONE",
              "MHz", "REV", "HOME", "BAND", "VFO/MR"]
SQLPRESET = ["Off", "2", "5", "9", "Full"]
BANDS = ["30 MHz", "50 MHz", "60 MHz", "108 MHz", "150 MHz", "250 MHz", "350 MHz",
         "450 MHz", "850 MHz"]
STEPS = [2.5, 5.0, 6.25, 7.5, 8.33, 10.0, 12.5,
         15.0, 20.0, 25.0, 30.0, 50.0, 100.0]


def add_radio_setting(radio_setting_group, mem_field, ui_name, option_map,
                      current, doc=None):
    setting = RadioSetting(mem_field, ui_name,
                           RadioSettingValueMap(option_map, current))
    if doc is not None:
        setting.set_doc(doc)
    radio_setting_group.append(setting)


def add_radio_bool(radio_setting_group, mem_field, ui_name, current, doc=None):
    setting = RadioSetting(mem_field, ui_name,
                           RadioSettingValueBoolean(bool(current)))
    radio_setting_group.append(setting)


class TYTTH7800Base(chirp_common.Radio):
    """Base class for TYT TH-7800"""
    VENDOR = "TYT"

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (1, 800)
        rf.has_bank = False
        rf.has_tuning_step = True
        rf.valid_tuning_steps = STEPS
        rf.can_odd_split = True
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_tmodes = TMODES
        rf.has_ctone = False
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC + "#*-+"
        rf.valid_bands = [(108000000, 180000000),
                          (350000000, 399995000),
                          (400000000, 512000000)]
        rf.valid_skips = SCAN_MODES
        rf.valid_modes = MODES + ["AM"]
        rf.valid_name_length = 6
        rf.has_settings = True
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(
            TH7800_MEM_FORMAT %
            (self._mmap_offset, self._scanlimits_offset, self._settings_offset,
             self._chan_active_offset, self._info_offset), self._mmap)

    def get_active(self, banktype, num):
        """get active flag for channel active,
        scan enable, or priority banks"""
        bank = getattr(self._memobj, banktype)
        index = (num - 1) // 8
        bitpos = (num - 1) % 8
        mask = 2**bitpos
        enabled = bank[index] & mask
        if enabled:
            return True
        else:
            return False

    def set_active(self, banktype, num, enable=True):
        """set active flag for channel active,
        scan enable, or priority banks"""
        bank = getattr(self._memobj, banktype)
        index = (num - 1) // 8
        bitpos = (num - 1) % 8
        mask = 2**bitpos
        if enable:
            bank[index] |= mask
        else:
            bank[index] &= ~mask

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def get_memory(self, number):
        _mem = self._memobj.memory[number - 1]
        mem = chirp_common.Memory()
        mem.number = number

        mem.empty = not self.get_active("chan_active", number)
        if mem.empty:
            return mem

        mem.freq = int(_mem.rx_freq) * 10

        txfreq = int(_mem.tx_freq) * 10
        if txfreq == mem.freq:
            mem.duplex = ""
        elif txfreq == 0:
            mem.duplex = "off"
            mem.offset = 0
        elif abs(txfreq - mem.freq) > 70000000:
            mem.duplex = "split"
            mem.offset = txfreq
        elif txfreq < mem.freq:
            mem.duplex = "-"
            mem.offset = mem.freq - txfreq
        elif txfreq > mem.freq:
            mem.duplex = "+"
            mem.offset = txfreq - mem.freq

        mem.dtcs_polarity = DTCS_POLARITY[_mem.dtcs_pol]

        mem.tmode = TMODES[int(_mem.tmode)]
        mem.ctone = mem.rtone = int(_mem.ctcss) / 10.0
        mem.dtcs = int(_mem.dtcs)

        mem.name = str(_mem.name)
        mem.name = mem.name.replace("\xFF", " ").rstrip()

        if not self.get_active("scan_enable", number):
            mem.skip = "S"
        elif self.get_active("priority", number):
            mem.skip = "P"
        else:
            mem.skip = ""

        mem.mode = _mem.am and "AM" or MODES[int(_mem.fmdev)]

        mem.power = POWER_LEVELS[_mem.power]
        mem.tuning_step = STEPS[_mem.step]

        mem.extra = RadioSettingGroup("extra", "Extra")

        add_radio_setting(mem.extra, "display", "Display",
                          zero_indexed_seq_map(["Frequency", "Name"]),
                          _mem.display)
        add_radio_setting(mem.extra, "hsdtype", "HSD TYPE",
                          zero_indexed_seq_map(["OFF", "2TON", "5TON",
                                                "DTMF"]),
                          _mem.hsdtype)
        add_radio_bool(mem.extra, "clk_sft", "CLK-SFT", _mem.clk_sft)
        add_radio_bool(mem.extra, "compand", "Compand", _mem.compand,
                       doc="Compress Audio")
        add_radio_bool(mem.extra, "talkaround", "Talk Around", _mem.talkaround,
                       doc="Simplex mode when out of range of repeater")

        add_radio_bool(mem.extra, "scramb", "Scramble", _mem.scramb,
                       doc="Frequency inversion Scramble")
        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]

        _prev_active = self.get_active("chan_active", mem.number)
        self.set_active("chan_active", mem.number, not mem.empty)
        if mem.empty or not _prev_active:
            LOG.debug("initializing memory channel %d" % mem.number)
            _mem.set_raw(BLANK_MEMORY)

        if mem.empty:
            return

        _mem.rx_freq = mem.freq / 10
        if mem.duplex == "split":
            _mem.tx_freq = mem.offset / 10
        elif mem.duplex == "-":
            _mem.tx_freq = (mem.freq - mem.offset) / 10
        elif mem.duplex == "+":
            _mem.tx_freq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "off":
            _mem.tx_freq = 0
            _mem.offset = 0
        else:
            _mem.tx_freq = mem.freq / 10

        _mem.tmode = TMODES.index(mem.tmode)
        if mem.tmode == "TSQL" or mem.tmode == "DTCS":
            _mem.sqlmode = 1
        else:
            _mem.sqlmode = 0
        _mem.ctcss = mem.rtone * 10
        _mem.dtcs = mem.dtcs
        _mem.dtcs_pol = DTCS_POLARITY.index(mem.dtcs_polarity)

        _mem.name = mem.name.ljust(6, "\xFF")

        # autoset display to name if filled, else show frequency
        if mem.extra:
            # mem.extra only seems to be populated when called from edit panel
            display = mem.extra["display"]
        else:
            display = None
        if mem.name:
            _mem.display = True
        else:
            _mem.display = False

        _mem.scan = SCAN_MODES.index(mem.skip)
        if mem.skip == "P":
            self.set_active("priority", mem.number, True)
            self.set_active("scan_enable", mem.number, True)
        elif mem.skip == "S":
            self.set_active("priority", mem.number, False)
            self.set_active("scan_enable", mem.number, False)
        elif mem.skip == "":
            self.set_active("priority", mem.number, False)
            self.set_active("scan_enable", mem.number, True)

        if mem.mode == "AM":
            _mem.am = True
            _mem.fmdev = 0
        else:
            _mem.am = False
            _mem.fmdev = MODES.index(mem.mode)

        if mem.power:
            _mem.power = POWER_LEVELS.index(mem.power)
        else:
            _mem.power = 0    # low
        _mem.step = STEPS.index(mem.tuning_step)

        for setting in mem.extra:
            LOG.debug("@set_mem: %s %s", setting.get_name(), setting.value)
            setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _settings = self._memobj.settings
        _info = self._memobj.info
        _bandlimits = self._memobj.bandlimits
        basic = RadioSettingGroup("basic", "Basic")
        info = RadioSettingGroup("info", "Model Info")
        top = RadioSettings(basic, info)
        add_radio_bool(basic, "beep", "Beep", _settings.beep)
        add_radio_bool(basic, "ars", "Auto Repeater Shift", _settings.ars)
        add_radio_setting(basic, "keylock", "Key Lock",
                          zero_indexed_seq_map(["Manual", "Auto"]),
                          _settings.keylock)
        add_radio_bool(basic, "auto_am", "Auto AM", _settings.auto_am)
        add_radio_setting(basic, "left_sql", "Left Squelch",
                          zero_indexed_seq_map(SQLPRESET),
                          _settings.left_sql)
        add_radio_setting(basic, "right_sql", "Right Squelch",
                          zero_indexed_seq_map(SQLPRESET),
                          _settings.right_sql)
        add_radio_setting(basic, "apo", "Auto Power off (Hours)",
                          [("Off", 0), ("0.5", 5), ("1.0", 10), ("1.5", 15),
                           ("2.0", 20)],
                          _settings.apo)
        add_radio_setting(basic, "backlight", "Display Backlight",
                          zero_indexed_seq_map(["Off", "1", "2", "3", "Full"]),
                          _settings.backlight)
        add_radio_setting(basic, "pttlock", "PTT Lock",
                          zero_indexed_seq_map(["Off", "Right", "Left",
                                                "Both"]),
                          _settings.pttlock)
        add_radio_setting(basic, "hyper_chan", "Hyper Channel",
                          zero_indexed_seq_map(["Manual", "Auto"]),
                          _settings.hyper_chan)
        add_radio_setting(basic, "right_func_key", "Right Function Key",
                          zero_indexed_seq_map(["Key 1", "Key 2"]),
                          _settings.right_func_key)
        add_radio_setting(basic, "mute_mode", "Mute Mode",
                          zero_indexed_seq_map(["Off", "TX", "RX", "TX RX"]),
                          _settings.mute_mode)
        add_radio_setting(basic, "scan_mode", "Scan Mode",
                          zero_indexed_seq_map(["MEM", "MSM"]),
                          _settings.scan_mode,
                          doc="MEM = Normal scan, bypass channels marked "
                          "skip. MSM = Scan only channels marked priority.")
        add_radio_setting(basic, "scan_resume", "Scan Resume",
                          zero_indexed_seq_map(["Time", "Busy"]),
                          _settings.scan_resume)
        basic.append(RadioSetting(
                "tot", "Time Out Timer (minutes)",
                RadioSettingValueInteger(0, 30, _settings.tot)))
        add_radio_setting(basic, "p1", "P1 Function",
                          zero_indexed_seq_map(MICKEYFUNC),
                          _settings.p1)
        add_radio_setting(basic, "p2", "P2 Function",
                          zero_indexed_seq_map(MICKEYFUNC),
                          _settings.p2)
        add_radio_setting(basic, "p3", "P3 Function",
                          zero_indexed_seq_map(MICKEYFUNC),
                          _settings.p3)
        add_radio_setting(basic, "p4", "P4 Function",
                          zero_indexed_seq_map(MICKEYFUNC),
                          _settings.p4)

        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in chirp_common.CHARSET_ASCII:
                    filtered += char
                else:
                    filtered += " "
            return filtered

        rsvs = RadioSettingValueString(0, 8, _filter(_info.sn))
        rsvs.set_mutable(False)
        rs = RadioSetting("sn", "Serial Number", rsvs)
        info.append(rs)

        rsvs = RadioSettingValueString(0, 8, _filter(_info.model))
        rsvs.set_mutable(False)
        rs = RadioSetting("model", "Model Name", rsvs)
        info.append(rs)

        rsvs = RadioSettingValueString(0, 16, _filter(_info.code))
        rsvs.set_mutable(False)
        rs = RadioSetting("code", "Model Code", rsvs)
        info.append(rs)

        progdate = "%d/%d/%d" % (_info.prog_mon, _info.prog_day,
                                 _info.prog_yr)
        rsvs = RadioSettingValueString(0, 10, progdate)
        rsvs.set_mutable(False)
        rs = RadioSetting("progdate", "Last Program Date", rsvs)
        info.append(rs)

        # Band Limits
        for i in range(0, len(BANDS)):
            rx_start = int(_bandlimits[i].lorx) * 10
            if not rx_start == 0:
                objname = BANDS[i] + "lorx"
                objnamepp = BANDS[i] + " Rx Start"
                rsv = RadioSettingValueString(0, 10, format_freq(rx_start))
                rsv.set_mutable(False)
                rs = RadioSetting(objname, objnamepp, rsv)
                info.append(rs)

                rx_end = int(_bandlimits[i].hirx) * 10
                objname = BANDS[i] + "hirx"
                objnamepp = BANDS[i] + " Rx end"
                rsv = RadioSettingValueString(0, 10, format_freq(rx_end))
                rsv.set_mutable(False)
                rs = RadioSetting(objname, objnamepp, rsv)
                info.append(rs)

            tx_start = int(_bandlimits[i].lotx) * 10
            if not tx_start == 0:
                objname = BANDS[i] + "lotx"
                objnamepp = BANDS[i] + " Tx Start"
                rsv = RadioSettingValueString(0, 10, format_freq(tx_start))
                rsv.set_mutable(False)
                rs = RadioSetting(objname, objnamepp, rsv)
                info.append(rs)

                tx_end = int(_bandlimits[i].hitx) * 10
                objname = BANDS[i] + "hitx"
                objnamepp = BANDS[i] + " Tx end"
                rsv = RadioSettingValueString(0, 10, format_freq(tx_end))
                rsv.set_mutable(False)
                rs = RadioSetting(objname, objnamepp, rsv)
                info.append(rs)
        return top

    def set_settings(self, settings):
        _settings = self._memobj.settings
        _info = self._memobj.info
        _bandlimits = self._memobj.bandlimits
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue
            try:
                setting = element.get_name()
                oldval = getattr(_settings, setting)
                newval = element.value

                LOG.debug("Setting %s(%s) <= %s" % (setting, oldval, newval))
                setattr(_settings, setting, newval)
            except Exception:
                LOG.debug(element.get_name())
                raise


@directory.register
class TYTTH7800File(TYTTH7800Base, chirp_common.FileBackedRadio):
    """TYT TH-7800 .dat file"""
    MODEL = "TH-7800 File"

    FILE_EXTENSION = "dat"

    _memsize = 69632
    _mmap_offset = 0x1100
    _scanlimits_offset = 0xC800 + _mmap_offset
    _settings_offset = 0xCB20 + _mmap_offset
    _chan_active_offset = 0xCB80 + _mmap_offset
    _info_offset = 0xfe00 + _mmap_offset

    def __init__(self, pipe):
        self.errors = []
        self._mmap = None

        if isinstance(pipe, str):
            self.pipe = None
            self.load_mmap(pipe)
        else:
            chirp_common.FileBackedRadio.__init__(self, pipe)

    @classmethod
    def match_model(cls, filedata, filename):
        return len(filedata) == cls._memsize and filename.endswith('.dat')


def _identify(radio):
    """Do identify handshake with TYT"""
    try:
        radio.pipe.write(b"\x02SPECPR")
        ack = radio.pipe.read(1)
        if ack != b"A":
            util.hexprint(ack)
            raise errors.RadioError("Radio did not ACK first command: %r"
                                    % ack)
    except:
        raise errors.RadioError("Unable to communicate with the radio")

    radio.pipe.write(b"G\x02")
    ident = radio.pipe.read(16)
    radio.pipe.write(b"A")
    r = radio.pipe.read(2)
    if r != b"A":
        raise errors.RadioError("Ack failed")
    return ident


def _download(radio, memsize=0x10000, blocksize=0x80):
    """Download from TYT TH-7800"""
    data = _identify(radio)
    LOG.info("ident:", util.hexprint(data))
    offset = 0x100
    for addr in range(offset, memsize, blocksize):
        msg = struct.pack(">cHB", b"R", addr, blocksize)
        radio.pipe.write(msg)
        block = radio.pipe.read(blocksize + 4)
        if len(block) != (blocksize + 4):
            LOG.debug(util.hexprint(block))
            raise errors.RadioError("Radio sent a short block")
        radio.pipe.write(b"A")
        ack = radio.pipe.read(1)
        if ack != b"A":
            LOG.debug(util.hexprint(ack))
            raise errors.RadioError("Radio NAKed block")
        data += block[4:]

        if radio.status_fn:
            status = chirp_common.Status()
            status.cur = addr
            status.max = memsize
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    radio.pipe.write(b"ENDR")

    return memmap.MemoryMapBytes(data)


def _upload(radio, memsize=0xF400, blocksize=0x80):
    """Upload to TYT TH-7800"""
    data = _identify(radio)

    radio.pipe.timeout = 1

    if data != radio._mmap[0:radio._mmap_offset]:
        raise errors.RadioError(
            "Model mismatch: \n%s\n%s" %
            (util.hexprint(data),
             util.hexprint(radio._mmap[:radio._mmap_offset])))
    # in the factory software they update the last program date when
    # they upload, So let's do the same
    today = date.today()
    y = today.year
    m = today.month
    d = today.day
    _info = radio._memobj.info

    ly = _info.prog_yr
    lm = _info.prog_mon
    ld = _info.prog_day
    LOG.debug("Updating last program date:%d/%d/%d" % (lm, ld, ly))
    LOG.debug("                  to today:%d/%d/%d" % (m, d, y))

    _info.prog_yr = y
    _info.prog_mon = m
    _info.prog_day = d

    offset = 0x0100
    for addr in range(offset, memsize, blocksize):
        mapaddr = addr + radio._mmap_offset - offset
        LOG.debug("addr: 0x%04X, mmapaddr: 0x%04X" % (addr, mapaddr))
        msg = struct.pack(">cHB", b"W", addr, blocksize)
        msg += radio._mmap[mapaddr:(mapaddr + blocksize)]
        LOG.debug(util.hexprint(msg))
        radio.pipe.write(msg)
        ack = radio.pipe.read(1)
        if ack != b"A":
            LOG.debug(util.hexprint(ack))
            raise errors.RadioError("Radio did not ack block 0x%04X" % addr)

        if radio.status_fn:
            status = chirp_common.Status()
            status.cur = addr
            status.max = memsize
            status.msg = "Cloning to radio"
            radio.status_fn(status)

    # End of clone
    radio.pipe.write(b"ENDW")

    # Checksum?
    final_data = radio.pipe.read(3)
    LOG.debug("final:", util.hexprint(final_data))


@directory.register
class TYTTH7800Radio(TYTTH7800Base, chirp_common.CloneModeRadio,
                     chirp_common.ExperimentalRadio):
    VENDOR = "TYT"
    MODEL = "TH-7800"
    BAUD_RATE = 38400
    NEEDS_COMPAT_SERIAL = False

    _memsize = 65296
    _mmap_offset = 0x0010
    _scanlimits_offset = 0xC800 + _mmap_offset
    _settings_offset = 0xCB20 + _mmap_offset
    _chan_active_offset = 0xCB80 + _mmap_offset
    _info_offset = 0xfe00 + _mmap_offset

    @classmethod
    def match_model(cls, filedata, filename):
        if len(filedata) != cls._memsize:
            return False
        # TYT used TH9800 as model for TH-7800 _AND_ TH-9800.  Check
        # for TH7800 in case they fix it or if users update the model
        # in their own radio.
        if not (filedata[0xfe18:0xfe1e] == b"TH9800" or
                filedata[0xfe18:0xfe1e] == b"TH7800"):
            return False
        # TH-7800 bandlimits differ from TH-9800.  First band Invalid
        # (zero).
        first_bandlimit = struct.unpack("BBBBBBBBBBBBBBBB",
                                        filedata[0xfe40:0xfe50])
        if not all(v == 0 for v in first_bandlimit):
            return False
        return True

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = (
         'This is experimental support for TH-7800 '
         'which is still under development.\n'
         'Please ensure you have a good backup with OEM software.\n'
         'Also please send in bug and enhancement requests!\n'
         'You have been warned. Proceed at your own risk!')
        return rp

    def sync_in(self):
        try:
            self._mmap = _download(self)
        except Exception as e:
            raise errors.RadioError(
                    "Failed to communicate with the radio: %s" % e)
        self.process_mmap()

    def sync_out(self):
        try:
            _upload(self)
        except Exception as e:
            raise errors.RadioError(
                    "Failed to communicate with the radio: %s" % e)



# Copyright 2023:
# * Sander van der Wel, <svdwel@icloud.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp import chirp_common, directory
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettings, RadioSettingValueString
import struct
from chirp.drivers import baofeng_common, baofeng_uv17Pro
from chirp import errors, util

LOG = logging.getLogger(__name__)
LIST_DTMFST = ["Off", "DT-ST", "ANI-ST", "DT+ANI"]
LIST_SAVE = ["Off", "1:1", "1:2", "1:3", "1:4"]
LIST_SCANMODE = ["Time", "Carrier", "Search"]


# Radios seem to have different memory sizes
def _get_memory_size(radio):
    response = baofeng_uv17Pro._sendmagic(radio, radio._magic_memsize[0][0],
                                          radio._magic_memsize[0][1])
    mem_size = struct.unpack("<I", response[7:])[0]
    for magic, resplen in radio._magics2:
        baofeng_uv17Pro._sendmagic(radio, magic, resplen)
    return mem_size


# Locations of memory may differ each time, so a mapping has to be made first
def _get_memory_map(radio):
    # Get memory map
    memory_map = []
    mem_size = _get_memory_size(radio)
    if mem_size != radio._radio_memsize:
        raise errors.RadioError("Incorrect radio model or model not supported "
                                "(memory size doesn't match)")
    for addr in range(0x1FFF, mem_size + 0x1000, 0x1000):
        frame = radio._make_frame(b"R", addr, 1)
        baofeng_common._rawsend(radio, frame)
        blocknr = ord(baofeng_common._rawrecv(radio, 6)[5:])
        blocknr = (blocknr >> 4 & 0xf) * 10 + (blocknr & 0xf)
        memory_map += [blocknr]
        baofeng_uv17Pro._sendmagic(radio, b"\x06", 1)
    return memory_map


def _download(radio):
    """Get the memory map"""
    baofeng_uv17Pro._do_ident(radio)
    data = b""
    memory_map = _get_memory_map(radio)

    status = chirp_common.Status()
    status.cur = 0
    status.max = radio.MEM_TOTAL // radio.BLOCK_SIZE
    status.msg = "Cloning from radio..."
    radio.status_fn(status)

    for block_number in radio.BLOCK_ORDER:
        if block_number not in memory_map:
            # Memory block not found.
            raise errors.RadioError('Radio memory is corrupted. ' +
                                    'Fix this by uploading a backup image ' +
                                    'to the radio.')
        block_index = memory_map.index(block_number) + 1
        start_addr = block_index * 0x1000
        for addr in range(start_addr, start_addr + 0x1000,
                          radio.BLOCK_SIZE):
            frame = radio._make_read_frame(addr, radio.BLOCK_SIZE)
            # DEBUG
            LOG.debug("Frame=" + util.hexprint(frame))

            baofeng_common._rawsend(radio, frame)

            d = baofeng_common._rawrecv(radio, radio.BLOCK_SIZE + 5)

            LOG.debug("Response Data= " + util.hexprint(d))

            data += d[5:]

            status.cur = len(data) // radio.BLOCK_SIZE
            status.msg = "Cloning from radio..."
            radio.status_fn(status)

            baofeng_uv17Pro._sendmagic(radio, b"\x06", 1)
    return data


def _upload(radio):
    """Upload procedure"""
    baofeng_uv17Pro._do_ident(radio)
    memory_map = _get_memory_map(radio)

    status = chirp_common.Status()
    status.cur = 0
    status.max = radio.WRITE_MEM_TOTAL // radio.BLOCK_SIZE
    status.msg = "Cloning to radio..."
    radio.status_fn(status)

    for block_number in radio.BLOCK_ORDER:
        # Choose a block number if memory map is corrupt
        # This can happen when the upload process was interrupted
        if block_number not in memory_map:
            memory_map[memory_map.index(165)] = block_number
        block_index = memory_map.index(block_number) + 1
        start_addr = block_index * 0x1000
        data_start_addr = radio.BLOCK_ORDER.index(block_number) * 0x1000
        for addr in range(start_addr, start_addr + 0x1000,
                          radio.BLOCK_SIZE):

            # sending the data
            data_addr = data_start_addr + addr - start_addr
            data = radio.get_mmap()[data_addr:data_addr + radio.BLOCK_SIZE]
            frame = radio._make_frame(b"W", addr, radio.BLOCK_SIZE, data)
            baofeng_common._rawsend(radio, frame)

            # receiving the response
            ack = baofeng_common._rawrecv(radio, 1)
            if ack != b"\x06":
                msg = "Bad ack writing block 0x%04x" % addr
                raise errors.RadioError(msg)

            # UI Update
            status.cur = (data_addr) // radio.BLOCK_SIZE
            status.msg = "Cloning to radio..."
            radio.status_fn(status)


@directory.register
class UV17(baofeng_uv17Pro.UV17Pro):
    """Baofeng UV-17"""
    VENDOR = "Baofeng"
    MODEL = "UV-17"
    NEEDS_COMPAT_SERIAL = False

    download_function = _download
    upload_function = _upload

    MODES = ["FM", "NFM"]
    BLOCK_ORDER = [16, 17, 18, 19,  24, 25, 26, 4, 6]
    MEM_TOTAL = 0x9000
    WRITE_MEM_TOTAL = 0x9000
    BLOCK_SIZE = 0x40
    BAUD_RATE = 57600
    _magic = b"PSEARCH"
    _magics = [(b"PASSSTA", 3),
               (b"SYSINFO", 1),
               (b"\x56\x00\x00\x0A\x0D", 13),
               (b"\x06", 1),
               (b"\x56\x00\x10\x0A\x0D", 13),
               (b"\x06", 1),
               (b"\x56\x00\x20\x0A\x0D", 13),
               (b"\x06", 1)]
    _magic_memsize = [(b"\x56\x00\x00\x00\x0A", 11)]
    _radio_memsize = 0xffff
    _magics2 = [(b"\x06", 1),
                (b"\xFF\xFF\xFF\xFF\x0C\x55\x56\x31\x35\x39\x39\x39", 1),
                (b"\02", 8),
                (b"\x06", 1)]
    _fingerprint = b"\x06" + b"UV15999"
    _scode_offset = 1

    _tri_band = False
    POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=1.00),
                    chirp_common.PowerLevel("High",  watts=5.00)]

    LENGTH_NAME = 11
    VALID_BANDS = [baofeng_uv17Pro.UV17Pro._vhf_range,
                   baofeng_uv17Pro.UV17Pro._uhf_range]
    SCODE_LIST = ["%s" % x for x in range(1, 16)]
    SQUELCH_LIST = ["Off"] + list("123456789")
    LIST_POWERON_DISPLAY_TYPE = ["Full", "Message", "Voltage"]
    LIST_TIMEOUT = ["Off"] + ["%s sec" % x for x in range(15, 615, 15)]
    LIST_VOICE = ["Chinese", "English"]
    LIST_BACKLIGHT_TIMER = ["Always On"] + ["%s sec" % x for x in range(1, 11)]
    LIST_MODE = ["Name", "Frequency"]
    CHANNELS = 999

    MEM_FORMAT = """
    #seekto 0x0030;
    struct channel {
      lbcd rxfreq[4];
      lbcd txfreq[4];
      u8 unused1;
      ul16 rxtone;
      ul16 txtone;
      u8 unknown1:1,
         bcl:1,
         pttid:2,
         unknown2:1,
         wide:1,
         lowpower:1,
         unknown:1;
      u8 scode:4,
         unknown3:3,
         scan:1;
      u8 unknown4;
    };

    struct {
      struct channel mem[252];
    } mem1;

    #seek 0x10;
    struct {
      struct channel mem[255];
    } mem2;

    #seek 0x10;
    struct {
      struct channel mem[255];
    } mem3;

    #seek 0x10;
    struct {
      struct channel mem[237];
    } mem4;

    #seekto 0x7000;
    struct {
      u8 powerondistype;
      u8 unknown0[15];
      char boottext1[10];
      u8 unknown1[6];
      char boottext2[10];
      u8 unknown2[22];
      u8 tot;
      u8 squelch;
      u8 vox;
      u8 powersave: 4,
         unknown3:2,
         voice: 1,
         voicesw: 1;
      u8 backlight;
      u8 beep:1,
         autolock:1,
         unknown4:1,
         tail:1,
         scanmode:2,
         dtmfst:2;
      u8 unknown5:1,
         dualstandby:1,
         roger:1,
         unknown6:3,
         fmenable:1,
         unknown7:1;
      u8 unknown8[9];
      u8 unknown9:6,
         chbdistype:1,
         chadistype:1;
    } settings;

    struct vfo {
      lbcd rxfreq[4];
      lbcd txfreq[4];
      u8 unused1;
      ul16 rxtone;
      ul16 txtone;
      u8 unknown1:1,
         bcl:1,
         pttid:2,
         unknown2:1,
         wide:1,
         lowpower:1,
         unknown:1;
      u8 scode:4,
         unknown3:3,
         scan:1;
      u8 unknown4;
    };

    #seekto 0x0010;
    struct {
      struct vfo a;
      struct vfo b;
    } vfo;

    #seekto 0x4000;
    struct {
      char name[11];
    } names[999];

    #seekto 0x8000;
    struct {
      u8 code[5];
    } pttid[15];

    struct {
      u8 unknown[5];
      u8 code[5];
    } ani;

    """

    def _make_frame(self, cmd, addr, length, data=""):
        """Pack the info in the header format"""
        frame = struct.pack("<cI", cmd, addr)[:-1] + struct.pack(">B", length)
        # add the data if set
        if len(data) != 0:
            frame += data
        # return the data
        return frame

    def get_settings(self):
        """Translate the bit in the mem_struct into settings in the UI"""
        _mem = self._memobj
        basic = RadioSettingGroup("basic", "Basic Settings")
        dtmfe = RadioSettingGroup("dtmfe", "DTMF Encode Settings")
        top = RadioSettings(basic, dtmfe)

        self.get_settings_common_basic(basic, _mem)

        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in chirp_common.CHARSET_ASCII:
                    filtered += char
                else:
                    filtered += " "
            return filtered

        rs = RadioSetting("settings.boottext1", "Power-On Message 1",
                          RadioSettingValueString(
                              0, 10, _filter(self._memobj.settings.boottext1)))
        basic.append(rs)

        rs = RadioSetting("settings.boottext2", "Power-On Message 2",
                          RadioSettingValueString(
                              0, 10, _filter(self._memobj.settings.boottext2)))
        basic.append(rs)

        if _mem.settings.powersave > 0x04:
            val = 0x00
        else:
            val = _mem.settings.powersave
        rs = RadioSetting("settings.powersave", "Battery Saver",
                          RadioSettingValueList(
                              LIST_SAVE, LIST_SAVE[val]))
        basic.append(rs)

        rs = RadioSetting("settings.scanmode", "Scan Mode",
                          RadioSettingValueList(
                              LIST_SCANMODE,
                              LIST_SCANMODE[_mem.settings.scanmode]))
        basic.append(rs)

        rs = RadioSetting("settings.dtmfst", "DTMF Sidetone",
                          RadioSettingValueList(LIST_DTMFST, LIST_DTMFST[
                              _mem.settings.dtmfst]))
        basic.append(rs)

        rs = RadioSetting("settings.fmenable", "Enable FM radio",
                          RadioSettingValueBoolean(_mem.settings.fmenable))
        basic.append(rs)

        self.get_settings_common_dtmf(dtmfe, _mem)

        return top

    def decode_tone(self, val):
        pol = "N"
        mode = ""
        if val in [0, 0xFFFF]:
            xval = 0
        elif (val & 0x8000) > 0:
            mode = "DTCS"
            xval = (val & 0x0f) + (val >> 4 & 0xf)\
                * 10 + (val >> 8 & 0xf) * 100
            if (val & 0xC000) == 0xC000:
                pol = "R"
        else:
            mode = "Tone"
            xval = int((val & 0x0f) + (val >> 4 & 0xf) * 10 +
                       (val >> 8 & 0xf) * 100 + (val >> 12 & 0xf)
                       * 1000) / 10.0

        return mode, xval, pol

    def get_raw_memory(self, number):
        # The flash memory contains page_numbers
        # This is probably to do wear leveling on the memory
        # These numbers are needed, but make the channel memory
        # not continuous.
        number = number - 1
        if number >= 762:
            _mem = self._memobj.mem4.mem[number - 762]
            return _mem
        if number >= 507:
            _mem = self._memobj.mem3.mem[number - 507]
            return _mem
        if number >= 252:
            _mem = self._memobj.mem2.mem[number - 252]
            return _mem
        _mem = self._memobj.mem1.mem[number]
        return _mem

    def get_memory(self, number):
        _mem = self.get_raw_memory(number)
        _nam = self._memobj.names[number - 1]

        mem = chirp_common.Memory()
        mem.number = number

        self.get_memory_common(_mem, _nam.name, mem)

        return mem

    def encode_tone(self, memtone, mode, tone, pol):
        if mode == "Tone":
            xtone = '%04i' % (tone * 10)
            memtone.set_value((int(xtone[0]) << 12) + (int(xtone[1]) << 8) +
                              (int(xtone[2]) << 4) + int(xtone[3]))
        elif mode == "TSQL":
            xtone = '%04i' % (tone * 10)
            memtone.set_value((int(tone[0]) << 12) + (int(xtone[1]) << 8) +
                              (int(xtone[2]) << 4) + int(xtone[3]))
        elif mode == "DTCS":
            xtone = str(int(tone)).rjust(4, '0')
            memtone.set_value((0x8000 + (int(xtone[0]) << 12) +
                               (int(xtone[1]) << 8) + (int(xtone[2]) << 4) +
                               int(xtone[3])))
        else:
            memtone.set_value(0)

        if mode == "DTCS" and pol == "R":
            memtone.set_value(memtone + 0x4000)

    def set_memory(self, mem):
        _mem = self.get_raw_memory(mem.number)
        _nam = self._memobj.names[mem.number - 1]

        if mem.empty:
            _mem.set_raw(b"\xff" * 16)
            return

        _mem.set_raw(b"\x00" * 16)

        _namelength = self.get_features().valid_name_length
        _nam.name = mem.name.ljust(_namelength, '\x00')

        self.set_memory_common(mem, _mem)


@directory.register
class UV13Pro(UV17):
    VENDOR = "Baofeng"
    MODEL = "UV-13Pro"

    _radio_memsize = 0x31fff



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import chirp_common, errors, util
from chirp.drivers import tmv71_ll
import logging

LOG = logging.getLogger(__name__)


class TMV71ARadio(chirp_common.CloneModeRadio):
    BAUD_RATE = 9600
    VENDOR = "Kenwood"
    MODEL = "TM-V71A"

    mem_upper_limit = 1022
    _memsize = 32512
    _model = ""  # FIXME: REMOVE

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (0, 999)
        return rf

    def _detect_baud(self):
        for baud in [9600, 19200, 38400, 57600]:
            self.pipe.baudrate = baud
            self.pipe.write("\r\r")
            self.pipe.read(32)
            try:
                id = tmv71_ll.get_id(self.pipe)
                LOG.info("Radio %s at %i baud" % (id, baud))
                return True
            except errors.RadioError:
                pass

        raise errors.RadioError("No response from radio")

    def get_raw_memory(self, number):
        return util.hexprint(tmv71_ll.get_raw_mem(self._mmap, number))

    def get_special_locations(self):
        return sorted(tmv71_ll.V71_SPECIAL.keys())

    def get_memory(self, number):
        if isinstance(number, str):
            try:
                number = tmv71_ll.V71_SPECIAL[number]
            except KeyError:
                raise errors.InvalidMemoryLocation("Unknown channel %s" %
                                                   number)

        return tmv71_ll.get_memory(self._mmap, number)

    def set_memory(self, mem):
        return tmv71_ll.set_memory(self._mmap, mem)

    def erase_memory(self, number):
        tmv71_ll.set_used(self._mmap, number, 0)

    def sync_in(self):
        self._detect_baud()
        self._mmap = tmv71_ll.download(self)

    def sync_out(self):
        self._detect_baud()
        tmv71_ll.upload(self)



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import serial
import logging

from chirp import chirp_common, errors, directory
from chirp.drivers import ic9x_ll, icf, kenwood_live, icomciv

LOG = logging.getLogger(__name__)


class DetectorRadio(chirp_common.Radio):
    """Minimal radio for model detection"""

    def get_payload(self, data, raw, checksum):
        return data


def _icom_model_data_to_rclass(md):
    for _rtype, rclass in list(directory.DRV_TO_RADIO.items()):
        if rclass.VENDOR != "Icom":
            continue
        if not hasattr(rclass, 'get_model') or not rclass.get_model():
            continue
        if rclass.get_model()[:4] == md[:4]:
            return rclass

    raise errors.RadioError("Unknown radio type %02x%02x%02x%02x" %
                            (ord(md[0]), ord(md[1]), ord(md[2]), ord(md[3])))


def _detect_icom_radio(ser):
    # ICOM VHF/UHF Clone-type radios @ 9600 baud

    try:
        ser.baudrate = 9600
        md = icf.get_model_data(DetectorRadio(ser))
        return _icom_model_data_to_rclass(md)
    except errors.RadioError as e:
        LOG.error("_detect_icom_radio: %s", e)

    # ICOM IC-91/92 Live-mode radios @ 4800/38400 baud

    ser.baudrate = 4800
    try:
        ic9x_ll.send_magic(ser)
        return _icom_model_data_to_rclass("ic9x")
    except errors.RadioError:
        pass

    # ICOM CI/V Radios @ various bauds

    for rate in [9600, 4800, 19200]:
        try:
            ser.baudrate = rate
            return icomciv.probe_model(ser)
        except errors.RadioError:
            pass

    ser.close()

    raise errors.RadioError("Unable to get radio model")


def detect_icom_radio(port):
    """Detect which Icom model is connected to @port"""
    if '://' in port:
        ser = serial.serial_for_url(port, do_not_open=True)
        ser.timeout = 0.5
        ser.open()
    else:
        ser = serial.Serial(port=port, timeout=0.5)

    try:
        result = _detect_icom_radio(ser)
    except Exception:
        ser.close()
        raise

    ser.close()

    LOG.info("Auto-detected %s %s on %s" %
             (result.VENDOR, result.MODEL, port))

    return result


def detect_kenwoodlive_radio(port):
    """Detect which Kenwood model is connected to @port"""
    if '://' in port:
        ser = serial.serial_for_url(port, do_not_open=True)
        ser.timeout = 0.5
        ser.open()
    else:
        ser = serial.Serial(port=port, baudrate=9600, timeout=0.5)

    r_id = kenwood_live.get_id(ser)
    ser.close()

    models = {}
    for rclass in list(directory.DRV_TO_RADIO.values()):
        if rclass.VENDOR == "Kenwood":
            models[rclass.MODEL] = rclass

    if r_id in list(models.keys()):
        return models[r_id]
    else:
        raise errors.RadioError("Unsupported model `%s'" % r_id)


DETECT_FUNCTIONS = {
    "Icom":    detect_icom_radio,
    "Kenwood": detect_kenwoodlive_radio,
}



# Copyright 2012 Tom Hayward <tom@tomh.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# FT-1802 Clone Procedure
# 1. Turn radio off.
# 2. Connect cable to mic jack.
# 3. Press and hold in the [LOW(A/N)] key while turning the radio on.
# 4. In Chirp, choose Download from Radio.
# 5. Press the [MHz(SET)] key to send image.
# or
# 4. Press the [D/MR(MW)] key ("--WAIT--" will appear on the LCD).
# 5. In Chirp, choose Upload to Radio.

from chirp.drivers import yaesu_clone
from chirp import chirp_common, bitwise, directory
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueBoolean

MEM_FORMAT = """
#seekto 0x06ea;
struct {
  u8 odd_pskip:1,
     odd_skip:1,
     odd_visible:1,
     odd_valid:1,
     even_pskip:1,
     even_skip:1,
     even_visible:1,
     even_valid:1;
} flags[100];

#seekto 0x076a;
struct {
  u8 unknown1a:1,
     step_changed:1,
     narrow:1,
     clk_shift:1,
     unknown1b:4;
  u8 unknown2a:2,
     duplex:2,
     unknown2b:1,
     tune_step:3;
  bbcd freq[3];
  u8 power:2,
     unknown3:3,
     tmode:3;
  u8 name[6];
  bbcd offset[3];
  u8 tone;
  u8 dtcs;
  u8 unknown4;
} memory[200];
"""


MODES = ["FM", "NFM"]
TMODES = ["", "Tone", "TSQL", "DTCS", "TSQL-R", "Cross"]
CROSS_MODES = ["DTCS->", "Tone->DTCS", "DTCS->Tone"]
DUPLEX = ["", "-", "+", "split"]
STEPS = [5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 50.0, 100.0]
POWER_LEVELS = [chirp_common.PowerLevel("LOW1", watts=5),
                chirp_common.PowerLevel("LOW2", watts=10),
                chirp_common.PowerLevel("LOW3", watts=25),
                chirp_common.PowerLevel("HIGH", watts=50),
                ]
CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ +-/?()?_"


@directory.register
class FT1802Radio(yaesu_clone.YaesuCloneModeRadio):
    """Yaesu FT-1802"""
    VENDOR = "Yaesu"
    MODEL = "FT-1802M"
    BAUD_RATE = 19200

    _model = b"AH023"
    _block_lengths = [10, 8001]
    _memsize = 8011

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable to mic jack.\n"
            "3. Press and hold in the [LOW(A/N)] key while turning the radio"
            " on.\n"
            "4. <b>After clicking OK</b>, press the [MHz(SET)] key to send"
            " image.\n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect cable to mic jack.\n"
            "3. Press and hold in the [LOW(A/N)] key while turning the radio"
            " on.\n"
            "4. Press the [D/MR(MW)] key (\"--WAIT--\" will appear on the"
            " LCD).\n")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()

        rf.memory_bounds = (0, 199)

        rf.can_odd_split = True
        rf.has_ctone = False
        rf.has_tuning_step = True
        rf.has_dtcs_polarity = False  # in radio settings, not per memory
        rf.has_bank = False  # has banks, but not implemented

        rf.valid_tuning_steps = STEPS
        rf.valid_modes = MODES
        rf.valid_tmodes = TMODES
        rf.valid_bands = [(137000000, 174000000)]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_duplexes = DUPLEX
        rf.valid_skips = ["", "S", "P"]
        rf.valid_name_length = 6
        rf.valid_characters = CHARSET
        rf.has_cross = True
        rf.valid_cross_modes = list(CROSS_MODES)

        return rf

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0, self._memsize-2)]

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number]) + \
               repr(self._memobj.flags[number/2])

    def get_memory(self, number):
        _mem = self._memobj.memory[number]
        _flag = self._memobj.flags[number/2]

        nibble = (number % 2) and "odd" or "even"
        visible = _flag["%s_visible" % nibble]
        valid = _flag["%s_valid" % nibble]
        pskip = _flag["%s_pskip" % nibble]
        skip = _flag["%s_skip" % nibble]

        mem = chirp_common.Memory()
        mem.number = number

        if not visible:
            mem.empty = True
        if not valid:
            mem.empty = True
            return mem

        mem.freq = chirp_common.fix_rounded_step(int(_mem.freq) * 1000)
        mem.offset = chirp_common.fix_rounded_step(int(_mem.offset) * 1000)
        mem.duplex = DUPLEX[_mem.duplex]
        mem.tuning_step = _mem.step_changed and \
            STEPS[_mem.tune_step] or STEPS[0]
        if _mem.tmode < TMODES.index("Cross"):
            mem.tmode = TMODES[_mem.tmode]
            mem.cross_mode = CROSS_MODES[0]
        else:
            mem.tmode = "Cross"
            mem.cross_mode = CROSS_MODES[_mem.tmode - TMODES.index("Cross")]
        mem.rtone = chirp_common.TONES[_mem.tone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        for i in _mem.name:
            if i == 0xFF:
                break
            if i & 0x80 == 0x80:
                # first bit in name is "show name"
                mem.name += CHARSET[0x80 ^ int(i)]
            else:
                mem.name += CHARSET[i]
        mem.name = mem.name.rstrip()
        mem.mode = _mem.narrow and "NFM" or "FM"
        mem.skip = pskip and "P" or skip and "S" or ""
        mem.power = POWER_LEVELS[_mem.power]

        mem.extra = RadioSettingGroup("extra", "Extra Settings")
        rs = RadioSetting("clk_shift", "Clock Shift",
                          RadioSettingValueBoolean(_mem.clk_shift))
        mem.extra.append(rs)

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number]
        _flag = self._memobj.flags[mem.number/2]

        nibble = (mem.number % 2) and "odd" or "even"

        valid = _flag["%s_valid" % nibble]
        visible = _flag["%s_visible" % nibble]

        if not mem.empty and not valid:
            _flag["%s_valid" % nibble] = True
            _mem.unknown1a = 0x00
            _mem.clk_shift = 0x00
            _mem.unknown1b = 0x00
            _mem.unknown2a = 0x00
            _mem.unknown2b = 0x00
            _mem.unknown3 = 0x00
            _mem.unknown4 = 0x00

        if mem.empty and valid and not visible:
            _flag["%s_valid" % nibble] = False
            return
        _flag["%s_visible" % nibble] = not mem.empty

        if mem.empty:
            return

        _flag["%s_valid" % nibble] = True

        _mem.freq = mem.freq / 1000
        _mem.offset = mem.offset / 1000
        _mem.duplex = DUPLEX.index(mem.duplex)
        _mem.tune_step = STEPS.index(mem.tuning_step)
        _mem.step_changed = mem.tuning_step != STEPS[0]
        if mem.tmode != "Cross":
            _mem.tmode = TMODES.index(mem.tmode)
        else:
            _mem.tmode = TMODES.index("Cross") + \
                         CROSS_MODES.index(mem.cross_mode)
        _mem.tone = chirp_common.TONES.index(mem.rtone)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)

        _mem.name = [0xFF] * 6
        for i in range(0, len(mem.name)):
            try:
                _mem.name[i] = CHARSET.index(mem.name[i])
            except IndexError:
                raise Exception("Character `%s' not supported")
        if _mem.name[0] != 0xFF:
            _mem.name[0] += 0x80  # show name instead of frequency

        _mem.narrow = MODES.index(mem.mode)
        _mem.power = 3 if mem.power is None else POWER_LEVELS.index(mem.power)

        _flag["%s_pskip" % nibble] = mem.skip == "P"
        _flag["%s_skip" % nibble] = mem.skip == "S"

        for element in mem.extra:
            setattr(_mem, element.get_name(), element.value)



# Copyright 2008 Dan Smith <dsmith@danplanet.com>
# Copyright 2012 Tom Hayward <tom@tomh.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import csv
import logging

from chirp import chirp_common, errors, directory
from chirp.drivers import generic_csv

LOG = logging.getLogger(__name__)


class OmittedHeaderError(Exception):
    """An internal exception to indicate that a header was omitted"""
    pass


@directory.register
class ITMRadio(generic_csv.CSVRadio):
    """Kenwood ITM format"""
    VENDOR = "Kenwood"
    MODEL = "ITM"
    FILE_EXTENSION = "itm"

    ATTR_MAP = {
        "CH":            (int,  "number"),
        "RXF":           (chirp_common.parse_freq, "freq"),
        "NAME":          (str,  "name"),
        }

    def _clean_duplex(self, headers, line, mem):
        try:
            txfreq = chirp_common.parse_freq(
                        generic_csv.get_datum_by_header(headers, line, "TXF"))
        except ValueError:
            mem.duplex = "off"
            return mem

        if mem.freq == txfreq:
            mem.duplex = ""
        elif txfreq:
            mem.duplex = "split"
            mem.offset = txfreq

        return mem

    def _clean_number(self, headers, line, mem):
        zone = int(generic_csv.get_datum_by_header(headers, line, "ZN"))
        mem.number = zone * 100 + mem.number
        return mem

    def _clean_tmode(self, headers, line, mem):
        rtone = eval(generic_csv.get_datum_by_header(headers, line, "TXSIG"))
        ctone = eval(generic_csv.get_datum_by_header(headers, line, "RXSIG"))

        if rtone:
            mem.tmode = "Tone"
        if ctone:
            mem.tmode = "TSQL"

        mem.rtone = rtone or 88.5
        mem.ctone = ctone or mem.rtone

        return mem

    def load(self, filename=None):
        if filename is None and self._filename is None:
            raise errors.RadioError("Need a location to load from")

        if filename:
            self._filename = filename

        self._blank()

        f = open(self._filename, "r")
        for line in f:
            if line.strip() == "// Conventional Data":
                break

        reader = csv.reader(f, delimiter=chirp_common.SEPCHAR, quotechar='"')

        good = 0
        lineno = 0
        for line in reader:
            lineno += 1
            if lineno == 1:
                header = line
                continue

            if len(line) == 0:
                # End of channel data
                break

            if len(header) > len(line):
                LOG.error("Line %i has %i columns, expected %i" %
                          (lineno, len(line), len(header)))
                self.errors.append("Column number mismatch on line %i" %
                                   lineno)
                continue

            # fix EU decimal
            line = [i.replace(',', '.') for i in line]

            try:
                mem = self._parse_csv_data_line(header, line)
                if mem.number is None:
                    raise Exception("Invalid Location field" % lineno)
            except Exception as e:
                LOG.error("Line %i: %s" % (lineno, e))
                self.errors.append("Line %i: %s" % (lineno, e))
                continue

            self._grow(mem.number)
            self.memories[mem.number] = mem
            good += 1

        if not good:
            for error in errors:
                LOG.error("kenwood_itm: %s", error)
            raise errors.InvalidDataError("No channels found")

    @classmethod
    def match_model(cls, filedata, filename):
        return filename.lower().endswith("." + cls.FILE_EXTENSION)



# Copyright 2011 Dan Smith <dsmith@danplanet.com>
#
# FT-2900-specific modifications by Richard Cochran, <ag6qr@sonic.net>
# Initial work on settings by Chris Fosnight, <chris.fosnight@gmail.com>
# FT-7100-specific modifications by Bruno Maire, <bruno@e48.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import logging

from chirp import util, memmap, chirp_common, bitwise, directory, errors
from chirp.drivers.yaesu_clone import YaesuCloneModeRadio
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList, RadioSettingValueString, RadioSettings, \
    RadioSettingValueInteger, RadioSettingValueBoolean


LOG = logging.getLogger(__name__)

ACK = b"\x06"
NB_OF_BLOCKS = 248
BLOCK_LEN = 32


def _send(pipe, data):
    time.sleep(0.035)   # Same delay as "FT7100 Programmer" from RT Systems
    # pipe.write(data) --> It seems, that the single bytes are sent too fast
    # so send character per character with a delay
    for ch in data:
        pipe.write(bytes([ch]))
        time.sleep(0.0012)  # 0.0011 is to short. No ACK after a few packets
    echo = pipe.read(len(data))
    if data == b"":
        raise Exception("Failed to read echo."
                        " Maybe serial hardware not connected."
                        " Maybe radio not powered or not in receiving mode.")
    if data != echo:
        LOG.debug("expecting echo\n%s\n", util.hexprint(data))
        LOG.debug("got echo\n%s\n", util.hexprint(echo))
        raise Exception("Got false echo. Expected: %r, got: %r.",
                        data, echo)


def _send_ack(pipe):
    time.sleep(0.01)  # Wait for radio input buffer ready
    # time.sleep(0.0003) is the absolute working minimum.
    # This delay is not critical for the transfer as there are not many ACKs.
    _send(pipe, ACK)


def _wait_for_ack(pipe):
    echo = pipe.read(1)
    if echo == b"":
        raise Exception("Failed to read ACK. No response from radio.")
    if echo != ACK:
        raise Exception("Failed to read ACK.  Expected: %r, got: %r.",
                        ACK, echo)


def _download(radio):
    LOG.debug("in _download\n")
    data = b""
    for _i in range(0, 60):
        chunk = radio.pipe.read(BLOCK_LEN)
        data += chunk
        LOG.debug("Header:\n%s", util.hexprint(data))
        if data == radio.IDBLOCK:
            break
        if len(data) > len(radio.IDBLOCK):
            break
    if data == b"":
        raise Exception("Got no data from radio.")
    if data != radio.IDBLOCK:
        raise Exception("Got false header. Expected: %r, got: %r." % (
                          radio.IDBLOCK, data))
    _send_ack(radio.pipe)

    # read 16 Byte block
    # and ignore it because it is constant. This might be a bug.
    # It was built in at the very beginning and discovered very late that the
    # data might be necessary later to write to the radio.
    # Now the data is hardcoded in _upload(radio)
    data = radio.pipe.read(16)
    _send_ack(radio.pipe)
    LOG.debug('Magic 16-byte chunk:\n%s' % util.hexprint(data))

    # initialize data, the big var that holds all memory
    data = b""
    for block_nr in range(NB_OF_BLOCKS):
        chunk = radio.pipe.read(BLOCK_LEN)
        if len(chunk) != BLOCK_LEN:
            LOG.debug("Block %i ", block_nr)
            LOG.debug("Got: %i:\n%s", len(chunk), util.hexprint(chunk))
            LOG.debug("len chunk is %i\n", len(chunk))
            raise Exception("Failed to get full data block")
        else:
            data += chunk
        _send_ack(radio.pipe)

        if radio.status_fn:
            status = chirp_common.Status()
            status.max = NB_OF_BLOCKS * BLOCK_LEN
            status.cur = len(data)
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    LOG.debug("Total: %i", len(data))
    _send_ack(radio.pipe)

    # for debugging purposes, dump the channels, in hex.
    for _i in range(0, (NB_OF_BLOCKS * BLOCK_LEN) // 26):
        _start_data = 4 + 26 * _i
        chunk = data[_start_data:_start_data + 26]
        LOG.debug("channel %i:\n%s", _i-21, util.hexprint(chunk))

    return memmap.MemoryMapBytes(data)


def _upload(radio):
    data = radio.pipe.read(256)  # Clear buffer
    _send(radio.pipe, radio.IDBLOCK)
    _wait_for_ack(radio.pipe)

    # write 16 Byte block
    # If there should be a problem, see remarks in _download(radio)
    _send(radio.pipe, b"\xEE\x77\x01\x00\x0E\x07\x0E\x07"
                      b"\x00\x00\x00\x00\x00\x02\x00\x00")
    _wait_for_ack(radio.pipe)

    for block_nr in range(NB_OF_BLOCKS):
        data = radio.get_mmap()[block_nr * BLOCK_LEN:
                                (block_nr + 1) * BLOCK_LEN]
        LOG.debug("Writing block_nr %i:\n%s", block_nr, util.hexprint(data))
        _send(radio.pipe, data)
        _wait_for_ack(radio.pipe)

        if radio.status_fn:
            status = chirp_common.Status()
            status.max = NB_OF_BLOCKS * BLOCK_LEN
            status.cur = block_nr * BLOCK_LEN
            status.msg = "Cloning to radio"
            radio.status_fn(status)
        block_nr += 1


MEM_FORMAT = """
struct mem {
  u8   is_used:1,
       is_masked:1,
       is_skip:1,
       unknown11:3,
       show_name:1,
       is_split:1;
  u8   unknown2;
  ul32 freq_rx_Hz;
  ul32 freq_tx_Hz;
  ul16 offset_10khz;
  u8   unknown_dependent_of_band_144_b0000_430_b0101:4,
       tuning_step_index_1_2:4;
  u8   unknown51:2,
       is_offset_minus:1,
       is_offset_plus:1,
       unknown52:1,
       tone_mode_index:3;
  u8   tone_index;
  u8   dtcs_index;
  u8   is_mode_am:1,
       unknown71:2,
       is_packet96:1,
       unknown72:2,
       power_index:2;
  u8   unknown81:2,
       tuning_step_index_2_2:4,
       unknown82:2;
  char name[6];
  u8   unknown9;
  u8   unknownA;
};

// Settings are often present multiple times.
// The memories which is written to are mapped here
struct
{
#seekto 0x41;
  u8        current_band;
#seekto 0xa1;
  u8        apo;
#seekto 0xa2;
  u8        ars_vhf;
#seekto 0xe2;
  u8        ars_uhf;
#seekto 0xa3;
  u8        arts_vhf;
#seekto 0xa3;
  u8        arts_uhf;
#seekto 0xa4;
  u8        beep;
#seekto 0xa5;
  u8        cwid;
#seekto 0x80;
  char        cwidw[6];
#seekto 0xa7;
  u8        dim;
#seekto 0xaa;
  u8        dcsnr_vhf;
#seekto 0xea;
  u8        dcsnr_uhf;
#seekto 0xab;
  u8        disp;
#seekto 0xac;
  u8        dtmfd;
#seekto 0xad;
  u8        dtmfs;
#seekto 0xae;
  u8        dtmfw;
#seekto 0xb0;
  u8        lockt;
#seekto 0xb1;
  u8        mic;
#seekto 0xb2;
  u8        mute;
#seekto 0xb4;
  u8        button[4];
#seekto 0xb8;
  u8        rf_sql_vhf;
#seekto 0xf8;
  u8        rf_sql_uhf;
#seekto 0xb9;
  u8        scan_vhf;
#seekto 0xf9;
  u8        scan_uhf;
#seekto 0xbc;
  u8        speaker_cnt;
#seekto 0xff;
  u8        tot;
#seekto 0xc0;
  u8        txnar_vhf;
#seekto 0x100;
  u8        txnar_uhf;
#seekto 0xc1;
  u8        vfotr;
#seekto 0xc2;
  u8        am;
} overlay;

// All known memories
#seekto 0x20;
  u8        nb_mem_used_vhf;
#seekto 0x22;
  u8        nb_mem_used_vhf_and_limits;
#seekto 0x24;
  u8        nb_mem_used_uhf;
#seekto 0x26;
  u8        nb_mem_used_uhf_and_limits;

#seekto 0x41;
  u8        current_band;

#seekto 0x42;
  u8        current_nb_mem_used_vhf_maybe_not;

#seekto 0x4c;
  u8        priority_channel_maybe_1;   // not_implemented
  u8        priority_channel_maybe_2;   // not_implemented
  u8        priority_channel;           // not_implemented

#seekto 0x87;
  u8        opt_01_apo_1_4;
#seekto 0xa1;
  u8        opt_01_apo_2_4;
#seekto 0xc5;
  u8        opt_01_apo_3_4;
#seekto 0xe1;
  u8        opt_01_apo_4_4;

#seekto 0x88;
  u8        opt_02_ars_vhf_1_2;
#seekto 0xa2;
  u8        opt_02_ars_vhf_2_2;
#seekto 0xc6;
  u8        opt_02_ars_uhf_1_2;
#seekto 0xe2;
  u8        opt_02_ars_uhf_2_2;

#seekto 0x89;
  u8        opt_03_arts_mode_vhf_1_2;
#seekto 0xa3;
  u8        opt_03_arts_mode_vhf_2_2;
#seekto 0xc7;
  u8        opt_03_arts_mode_uhf_1_2;
#seekto 0xa3;
  u8        opt_03_arts_mode_vhf_2_2;

#seekto 0x8a;
  u8        opt_04_beep_1_2;
#seekto 0xa4;
  u8        opt_04_beep_2_2;

#seekto 0x8b;
  u8        opt_05_cwid_on_1_4;
#seekto 0xa5;
  u8        opt_05_cwid_on_2_4;
#seekto 0xc9;
  u8        opt_05_cwid_on_3_4;
#seekto 0xe5;
  u8        opt_05_cwid_on_4_4;

#seekto 0x80;
  char        opt_06_cwidw[6];

#seekto 0x8d;
  u8        opt_07_dim_1_4;
#seekto 0xa7;
  u8        opt_07_dim_2_4;
#seekto 0xcb;
  u8        opt_07_dim_3_4;
#seekto 0xe7;
  u8        opt_07_dim_4_4;

#seekto 0x90;
  u8        opt_10_dcsnr_vhf_1_2;
#seekto 0xaa;
  u8        opt_10_dcsnr_vhf_2_2;
#seekto 0xce;
  u8        opt_10_dcsnr_uhf_1_2;
#seekto 0xea;
  u8        opt_10_dcsnr_uhf_2_2;

#seekto 0x91;
  u8        opt_11_disp_1_4;
#seekto 0xab;
  u8        opt_11_disp_2_4;
#seekto 0xcf;
  u8        opt_11_disp_3_4;
#seekto 0xeb;
  u8        opt_11_disp_4_4;

#seekto 0x92;
  u8        opt_12_dtmf_delay_1_4;
#seekto 0xac;
  u8        opt_12_dtmf_delay_2_4;
#seekto 0xd0;
  u8        opt_12_dtmf_delay_3_4;
#seekto 0xec;
  u8        opt_12_dtmf_delay_4_4;

#seekto 0x93;
  u8        opt_13_dtmf_speed_1_4;
#seekto 0xad;
  u8        opt_13_dtmf_speed_2_4;
#seekto 0xd1;
  u8        opt_13_dtmf_speed_3_4;
#seekto 0xed;
  u8        opt_13_dtmf_speed_4_4;

#seekto 0x94;
  u8        opt_14_dtmfw_index_1_4;
#seekto 0xae;
  u8        opt_14_dtmfw_index_2_4;
#seekto 0xd2;
  u8        opt_14_dtmfw_index_3_4;
#seekto 0xee;
  u8        opt_14_dtmfw_index_4_4;

#seekto 0x96;
  u8        opt_16_lockt_1_4;
#seekto 0xb0;
  u8        opt_16_lockt_2_4;
#seekto 0xd4;
  u8        opt_16_lockt_3_4;
#seekto 0xf0;
  u8        opt_16_lockt_4_4;

#seekto 0x97;
  u8        opt_17_mic_MH48_1_4;
#seekto 0xb1;
  u8        opt_17_mic_MH48_2_4;
#seekto 0xd5;
  u8        opt_17_mic_MH48_3_4;
#seekto 0xf1;
  u8        opt_17_mic_MH48_4_4;

#seekto 0x98;
  u8        opt_18_mute_1_4;
#seekto 0xb2;
  u8        opt_18_mute_2_4;
#seekto 0xd6;
  u8        opt_18_mute_3_4;
#seekto 0xf2;
  u8        opt_18_mute_4_4;

#seekto 0x9a;
  u8        opt_20_pg_p_1_4[4];
#seekto 0xb4;
  u8        opt_20_pg_p_2_4[4];
#seekto 0xd8;
  u8        opt_20_pg_p_3_4[4];
#seekto 0xf4;
  u8        opt_20_pg_p_4_4[4];

#seekto 0x9e;
  u8        opt_24_rf_sql_vhf_1_2;
#seekto 0xb8;
  u8        opt_24_rf_sql_vhf_2_2;
#seekto 0xdc;
  u8        opt_24_rf_sql_uhf_1_2;
#seekto 0xf8;
  u8        opt_24_rf_sql_uhf_2_2;

#seekto 0x9f;
  u8        opt_25_scan_resume_vhf_1_2;
#seekto 0xb9;
  u8        opt_25_scan_resume_vhf_2_2;
#seekto 0xdd;
  u8        opt_25_scan_resume_uhf_1_2;
#seekto 0xf9;
  u8        opt_25_scan_resume_uhf_2_2;

#seekto 0xbc;
  u8        opt_28_speaker_cnt_1_2;
#seekto 0xfc;
  u8        opt_28_speaker_cnt_2_2;

#seekto 0xbf;
  u8        opt_31_tot_1_2;
#seekto 0xff;
  u8        opt_31_tot_2_2;

#seekto 0xc0;
  u8        opt_32_tx_nar_vhf;
#seekto 0x100;
  u8        opt_32_tx_nar_uhf;

#seekto 0xc1;
  u8        opt_33_vfo_tr;

#seekto 0xc2;
  u8        opt_34_am_1_2;
#seekto 0x102;
  u8        opt_34_am_2_2;

#seekto 260;
struct {
  struct mem mem_struct;
  char        fill_ff[2];
} unknown00;

struct {
  struct mem mem_struct;
  char        fill_00[6];
} current_vfo_vhf_uhf[2];

struct {
  struct mem mem_struct;
  char        fill_ff[6];
} current_mem_vhf_uhf[2];

struct {
  struct mem mem_struct;
  char        fill_ff[6];
} home_vhf_uhf[2];

struct {
  struct mem mem_struct;
  char        fill_010003000000[6];
} vhome;

struct {
  struct mem mem_struct;
  char        fill_010001000000[6];
} unknown01;

struct {
  char name[32];
} Vertex_Standard_AH003M_Backup_DT;

struct mem
  memory[260];

struct {
  char name[24];
} Vertex_Standard_AH003M;

struct {
  u8 dtmf[16];
} dtmf_mem[16];
"""

MODES_VHF = ["FM", "AM"]
MODES_UHF = ["FM"]      # AM can be set but is ignored by the radio
DUPLEX = ["", "-", "+", "split"]
TONE_MODES_RADIO = ["", "Tone", "TSQL", "CTCSS Bell", "DTCS"]
TONE_MODES = ["", "Tone", "TSQL", "DTCS"]
POWER_LEVELS = [
    chirp_common.PowerLevel("Low", watts=5),
    chirp_common.PowerLevel("Low2", watts=10),
    chirp_common.PowerLevel("Low3", watts=20),
    chirp_common.PowerLevel("High", watts=35),
    ]
TUNING_STEPS = [5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 50.0]
SKIP_VALUES = ["", "S"]
CHARSET = r"!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^ _"
DTMF_CHARSET = "0123456789*# "
SPECIAL_CHANS = ['VFO-VHF', 'VFO-UHF', 'Home-VHF', 'Home-UHF', 'VFO', 'Home']
SCAN_LIMITS = ["L1", "U1", "L2", "U2", "L3", "U3", "L4", "U4", "L5", "U5"]


def do_download(radio):
    """This is your download function"""
    return _download(radio)


@directory.register
class FT7100Radio(YaesuCloneModeRadio):

    """Yaesu FT-7100M"""
    MODEL = "FT-7100M"
    VARIANT = ""
    IDBLOCK = b"Vartex Standard AH003M M-Map V04"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False

    # Return information about this radio's features, including
    # how many memories it has, what bands it supports, etc
    def get_features(self):
        LOG.debug("get_features")
        rf = chirp_common.RadioFeatures()
        rf.has_bank = False

        rf.memory_bounds = (0, 240)
        # This radio supports 120 + 10 + 120 + 10 = 260 memories
        # These are zero based for chirpc
        rf.valid_bands = [
            (108000000, 180000000),  # Supports 2-meters tx
            (320000000, 999990000),  # Supports 70-centimeters tx
            ]
        rf.can_odd_split = True
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.has_cross = False
        rf.has_dtcs_polarity = False
        rf.has_bank = False
        rf.has_bank_names = False
        rf.has_settings = True
        rf.has_sub_devices = True
        rf.valid_tuning_steps = TUNING_STEPS
        rf.valid_modes = MODES_VHF
        rf.valid_tmodes = TONE_MODES
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_duplexes = DUPLEX
        rf.valid_skips = SKIP_VALUES
        rf.valid_name_length = 6
        rf.valid_characters = CHARSET
        rf.valid_special_chans = SPECIAL_CHANS
        return rf

    def sync_in(self):
        start = time.time()
        try:
            self._mmap = _download(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        LOG.info("Downloaded in %.2f sec", (time.time() - start))
        self.process_mmap()

    def sync_out(self):
        self.pipe.timeout = 1
        start = time.time()
        try:
            _upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        LOG.info("Uploaded in %.2f sec", (time.time() - start))

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def get_memory(self, number):
        LOG.debug("get_memory Number: %r", number)

        # Create a high-level memory object to return to the UI
        mem = chirp_common.Memory()

        # Get a low-level memory object mapped to the image
        if isinstance(number, int) and number < 0:
            number = SPECIAL_CHANS[number + 10]
        if isinstance(number, str):
            mem.number = -10 + SPECIAL_CHANS.index(number)
            mem.extd_number = number
            band = 0
            if self._memobj.overlay.current_band != 0:
                band = 1
            if number == 'VFO-VHF':
                _mem = self._memobj.current_vfo_vhf_uhf[0].mem_struct
            elif number == 'VFO-UHF':
                _mem = self._memobj.current_vfo_vhf_uhf[1].mem_struct
            elif number == 'Home-VHF':
                _mem = self._memobj.home_vhf_uhf[0].mem_struct
            elif number == 'Home-UHF':
                _mem = self._memobj.home_vhf_uhf[1].mem_struct
            elif number == 'VFO':
                _mem = self._memobj.current_vfo_vhf_uhf[band].mem_struct
            elif number == 'Home':
                _mem = self._memobj.home_vhf_uhf[band].mem_struct
            _mem.is_used = True
        else:
            mem.number = number                 # Set the memory number
            _mem = self._memobj.memory[number]
            upper_channel = self._memobj.nb_mem_used_vhf
            upper_limit = self._memobj.nb_mem_used_vhf_and_limits
            if number >= upper_channel and number < upper_limit:
                i = number - upper_channel
                mem.extd_number = SCAN_LIMITS[i]
            if number >= 260-10:
                i = number - (260-10)
                mem.extd_number = SCAN_LIMITS[i]

        # Convert your low-level frequency to Hertz
        mem.freq = int(_mem.freq_rx_Hz)
        mem.name = str(_mem.name).rstrip()  # Set the alpha tag

        mem.rtone = chirp_common.TONES[_mem.tone_index]
        mem.ctone = chirp_common.TONES[_mem.tone_index]

        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs_index]
        mem.rx_dtcs = chirp_common.DTCS_CODES[_mem.dtcs_index]

        tmode_radio = TONE_MODES_RADIO[_mem.tone_mode_index]
        # CTCSS Bell is TSQL plus a flag in the extra setting
        is_ctcss_bell = tmode_radio == "CTCSS Bell"
        if is_ctcss_bell:
            mem.tmode = "TSQL"
        else:
            mem.tmode = tmode_radio

        mem.duplex = ""
        if _mem.is_offset_plus:
            mem.duplex = "+"
        elif _mem.is_offset_minus:
            mem.duplex = "-"
        elif _mem.is_split:
            mem.duplex = "split"

        if _mem.is_split:
            mem.offset = int(_mem.freq_tx_Hz)
        else:
            mem.offset = int(_mem.offset_10khz)*10000   # 10 kHz to Hz

        if _mem.is_mode_am:
            mem.mode = "AM"
        else:
            mem.mode = "FM"

        mem.power = POWER_LEVELS[_mem.power_index]

        mem.tuning_step = TUNING_STEPS[_mem.tuning_step_index_1_2]

        if _mem.is_skip:
            mem.skip = "S"
        else:
            mem.skip = ""

        # mem.comment = ""

        mem.empty = not _mem.is_used

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting("show_name", "Show Name",
                          RadioSettingValueBoolean(_mem.show_name))
        mem.extra.append(rs)
        rs = RadioSetting("is_masked", "Is Masked",
                          RadioSettingValueBoolean(_mem.is_masked))
        mem.extra.append(rs)
        rs = RadioSetting("is_packet96", "Packet 9600",
                          RadioSettingValueBoolean(_mem.is_packet96))
        mem.extra.append(rs)
        rs = RadioSetting("is_ctcss_bell", "CTCSS Bell",
                          RadioSettingValueBoolean(is_ctcss_bell))
        mem.extra.append(rs)

        return mem

    def set_memory(self, mem):
        LOG.debug("set_memory Number: %r", mem.number)
        if mem.number < 0:
            number = SPECIAL_CHANS[mem.number+10]
            if number == 'VFO-VHF':
                _mem = self._memobj.current_vfo_vhf_uhf[0].mem_struct
            elif number == 'VFO-UHF':
                _mem = self._memobj.current_vfo_vhf_uhf[1].mem_struct
            elif number == 'Home-VHF':
                _mem = self._memobj.home_vhf_uhf[0].mem_struct
            elif number == 'Home-UHF':
                _mem = self._memobj.home_vhf_uhf[1].mem_struct
            else:
                raise errors.RadioError("Unexpected Memory Number: %r",
                                        mem.number)
        else:
            _mem = self._memobj.memory[mem.number]

        _mem.name = mem.name.ljust(6)

        _mem.tone_index = chirp_common.TONES.index(mem.rtone)
        _mem.dtcs_index = chirp_common.DTCS_CODES.index(mem.dtcs)
        if mem.tmode == "TSQL":
            _mem.tone_index = chirp_common.TONES.index(mem.ctone)
        if mem.tmode == "DTSC-R":
            _mem.dtcs_index = chirp_common.DTCS_CODES.index(mem.rx_dtcs)

        _mem.is_offset_plus = 0
        _mem.is_offset_minus = 0
        _mem.is_split = 0
        _mem.freq_rx_Hz = mem.freq
        _mem.freq_tx_Hz = mem.freq
        if mem.duplex == "+":
            _mem.is_offset_plus = 1
            _mem.freq_tx_Hz = mem.freq + mem.offset
            _mem.offset_10khz = int(mem.offset/10000)
        elif mem.duplex == "-":
            _mem.is_offset_minus = 1
            _mem.freq_tx_Hz = mem.freq - mem.offset
            _mem.offset_10khz = int(mem.offset/10000)
        elif mem.duplex == "split":
            _mem.is_split = 1
            _mem.freq_tx_Hz = mem.offset
            # No change to _mem.offset_10khz

        _mem.is_mode_am = mem.mode == "AM"

        if mem.power:
            _mem.power_index = POWER_LEVELS.index(mem.power)

        _mem.tuning_step_index_1_2 = TUNING_STEPS.index(mem.tuning_step)

        _mem.is_skip = mem.skip == "S"

        _mem.is_used = not mem.empty

        # if name not empty: show name
        _mem.show_name = mem.name.strip() != ""

        # In testmode there is no setting in mem.extra
        # This is why the following line is not located in the else part of
        # the if structure below
        _mem.tone_mode_index = TONE_MODES_RADIO.index(mem.tmode)

        for setting in mem.extra:
            if setting.get_name() == "is_ctcss_bell":
                if mem.tmode == "TSQL" and setting.value:
                    _mem.tone_mode_index = TONE_MODES_RADIO.index("CTCSS Bell")
            else:
                setattr(_mem, setting.get_name(), setting.value)

        LOG.debug("encoded mem\n%s\n",
                  (util.hexprint(_mem.get_raw(asbytes=False)[0:25])))
        LOG.debug(repr(_mem))

    def get_settings(self):
        common = RadioSettingGroup("common", "Common Settings")
        band = RadioSettingGroup("band", "Band dependent Settings")
        arts = RadioSettingGroup("arts",
                                 "Auto Range Transponder System (ARTS)")
        dtmf = RadioSettingGroup("dtmf", "DTMF Settings")
        mic_button = RadioSettingGroup("mic_button", "Microphone Buttons")
        setmode = RadioSettings(common, band, arts, dtmf, mic_button)

        _overlay = self._memobj.overlay

        # numbers and names of settings refer to the way they're
        # presented in the set menu, as well as the list starting on
        # page 49 of the manual

        # 1 Automatic Power Off
        opts = [
            "Off", "30 Min",
            "1 Hour", "1.5 Hours",
            "2 Hours", "2.5 Hours",
            "3 Hours", "3.5 Hours",
            "4 Hours", "4.5 Hours",
            "5 Hours", "5.5 Hours",
            "6 Hours", "6.5 Hours",
            "7 Hours", "7.5 Hours",
            "8 Hours", "8.5 Hours",
            "9 Hours", "9.5 Hours",
            "10 Hours", "10.5 Hours",
            "11 Hours", "11.5 Hours",
            "12 Hours",
            ]
        common.append(
            RadioSetting(
                "apo", "Automatic Power Off",
                RadioSettingValueList(opts, opts[_overlay.apo])))

        # 2 Automatic Repeater Shift function
        opts = ["Off", "On"]
        band.append(
            RadioSetting(
                "ars_vhf", "Automatic Repeater Shift VHF",
                RadioSettingValueList(opts, opts[_overlay.ars_vhf])))
        band.append(
            RadioSetting(
                "ars_uhf", "Automatic Repeater Shift UHF",
                RadioSettingValueList(opts, opts[_overlay.ars_uhf])))

        # 3  Selects the ARTS mode.
        # -> Only useful to set it on the radio directly

        # 4 Enables/disables the key/button beeper.
        opts = ["Off", "On"]
        common.append(
            RadioSetting(
                "beep", "Key/Button Beep",
                RadioSettingValueList(opts, opts[_overlay.beep])))

        # 5 Enables/disables the CW IDer during ARTS operation.
        opts = ["Off", "On"]
        arts.append(
            RadioSetting(
                "cwid", "Enables/Disables the CW ID",
                RadioSettingValueList(opts, opts[_overlay.cwid])))

        # 6  Callsign during ARTS operation.
        cwidw = _overlay.cwidw.get_raw(asbytes=False)
        cwidw = cwidw.rstrip('\x00')
        val = RadioSettingValueString(0, 6, cwidw)
        val.set_charset(CHARSET)
        rs = RadioSetting("cwidw", "CW Identifier Callsign", val)

        def apply_cwid(setting):
            value_string = setting.value.get_value()
            _overlay.cwidw.set_value(value_string)
        rs.set_apply_callback(apply_cwid)
        arts.append(rs)

        # 7 Front panel display's illumination level.
        opts = ["0: Off", "1: Max", "2", "3", "4", "5", "6", "7: Min"]
        common.append(
            RadioSetting(
                "dim", "Display Illumination",
                RadioSettingValueList(opts, opts[_overlay.dim])))

        # 8 Setting the DCS code number.
        #   Note: This Menu item can be set independently for each band,
        #         and independently in each memory.

        # 9 Activates the DCS Code Search
        # -> Only useful if set on radio itself

        # 10 Selects 'Normal' or 'Inverted' DCS coding.
        opts = ["TRX Normal", "RX Reversed", "TX Reversed", "TRX Reversed"]
        band.append(
            RadioSetting(
                "dcsnr_vhf", "DCS coding VHF",
                RadioSettingValueList(opts, opts[_overlay.dcsnr_vhf])))
        band.append(
            RadioSetting(
                "dcsnr_uhf", "DCS coding UHF",
                RadioSettingValueList(opts, opts[_overlay.dcsnr_uhf])))

        # 11 Selects the 'sub' band display format
        opts = ["Frequency", "Off / Sub Band disabled",
                "DC Input Voltage", "CW ID"]
        common.append(
            RadioSetting(
                "disp", "Sub Band Display Format",
                RadioSettingValueList(opts, opts[_overlay.disp])))

        # 12 Setting the DTMF Autodialer delay time
        opts = ["50 ms", "250 ms", "450 ms", "750 ms", "1 s"]
        dtmf.append(
            RadioSetting(
                "dtmfd", "Autodialer delay time",
                RadioSettingValueList(opts, opts[_overlay.dtmfd])))

        # 13 Setting the DTMF Autodialer sending speed
        opts = ["50 ms", "75 ms", "100 ms"]
        dtmf.append(
            RadioSetting(
                "dtmfs", "Autodialer sending speed",
                RadioSettingValueList(opts, opts[_overlay.dtmfs])))

        # 14 Current DTMF Autodialer memory
        rs = RadioSetting("dtmfw", "Current Autodialer memory",
                          RadioSettingValueInteger(1, 16, _overlay.dtmfw + 1))

        def apply_dtmfw(setting):
            _overlay.dtmfw = setting.value.get_value() - 1
        rs.set_apply_callback(apply_dtmfw)
        dtmf.append(rs)

        # DTMF Memory
        for i in range(16):
            dtmf_string = ""
            for j in range(16):
                dtmf_char = ''
                dtmf_int = int(self._memobj.dtmf_mem[i].dtmf[j])
                if dtmf_int < 10:
                    dtmf_char = str(dtmf_int)
                elif dtmf_int == 14:
                    dtmf_char = '*'
                elif dtmf_int == 15:
                    dtmf_char = '#'
                elif dtmf_int == 255:
                    break
                dtmf_string += dtmf_char
            radio_setting_value_string = RadioSettingValueString(0, 16,
                                                                 dtmf_string)
            radio_setting_value_string.set_charset(DTMF_CHARSET)
            rs = RadioSetting("dtmf_%02i" % i,
                              "DTMF Mem " + str(i+1),
                              radio_setting_value_string)

            def apply_dtmf(setting, index):
                radio_setting_value_string = setting.value.get_value().rstrip()
                j = 0
                for dtmf_char in radio_setting_value_string:
                    dtmf_int = 255
                    if dtmf_char in "0123456789":
                        dtmf_int = int(dtmf_char)
                    elif dtmf_char == '*':
                        dtmf_int = 14
                    elif dtmf_char == '#':
                        dtmf_int = 15
                    if dtmf_int < 255:
                        self._memobj.dtmf_mem[index].dtmf[j] = dtmf_int
                        j += 1
                if j < 16:
                    self._memobj.dtmf_mem[index].dtmf[j] = 255
            rs.set_apply_callback(apply_dtmf, i)
            dtmf.append(rs)

        # 16 Enables/disables the PTT switch lock
        opts = ["Off", "Band A", "Band B", "Both"]
        common.append(
            RadioSetting(
                "lockt", "PTT switch lock",
                RadioSettingValueList(opts, opts[_overlay.lockt])))

        # 17 Selects the Microphone type to be used
        opts = ["MH-42", "MH-48"]
        common.append(
            RadioSetting(
                "mic", "Microphone type",
                RadioSettingValueList(opts, opts[_overlay.mic])))

        # 18 Reduces the audio level on the sub receiver when the
        #    main receiver is active
        opts = ["Off", "On"]
        common.append(
            RadioSetting(
                "mute", "Mute Sub Receiver",
                RadioSettingValueList(opts, opts[_overlay.mute])))

        # 20 - 23 Programming the microphones button assignment
        buttons = [
            "ACC / P1",
            "P / P2",
            "P1 / P3",
            "P2 / P4",
            ]
        opts_button = ["Low", "Tone", "MHz", "Rev", "Home", "Band",
                       "VFO / Memory", "Sql Off", "1750 Hz Tone Call",
                       "Repeater", "Priority"]
        for i, button in enumerate(buttons):
            rs = RadioSetting(
                "button" + str(i), button,
                RadioSettingValueList(opts_button,
                                      opts_button[_overlay.button[i]]))

            def apply_button(setting, index):
                value_string = setting.value.get_value()
                value_int = opts_button.index(value_string)
                _overlay.button[index] = value_int
            rs.set_apply_callback(apply_button, i)
            mic_button.append(rs)

        # 24 Adjusts the RF SQL threshold level
        opts = ["Off", "S-1", "S-5", "S-9", "S-FULL"]
        band.append(
            RadioSetting(
                "rf_sql_vhf", "RF Sql VHF",
                RadioSettingValueList(opts, opts[_overlay.rf_sql_vhf])))
        band.append(
            RadioSetting(
                "rf_sql_uhf", "RF Sql UHF",
                RadioSettingValueList(opts, opts[_overlay.rf_sql_uhf])))

        # 25 Selects the Scan-Resume mode
        opts = ["Busy", "Time"]
        band.append(
            RadioSetting(
                "scan_vhf", "Scan-Resume VHF",
                RadioSettingValueList(opts, opts[_overlay.scan_vhf])))
        band.append(
            RadioSetting(
                "scan_uhf", "Scan-Resume UHF",
                RadioSettingValueList(opts, opts[_overlay.scan_uhf])))

        # 28 Defining the audio path to the external speaker
        opts = ["Off", "Band A", "Band B", "Both"]
        common.append(
            RadioSetting(
                "speaker_cnt", "External Speaker",
                RadioSettingValueList(opts, opts[_overlay.speaker_cnt])))

        # 31 Sets the Time-Out Timer
        opts = ["Off", "Band A", "Band B", "Both"]
        common.append(
            RadioSetting(
                "tot", "TX Time-Out [Min.] (0 = Off)",
                RadioSettingValueInteger(0, 30, _overlay.tot)))

        # 32 Reducing the MIC Gain (and Deviation)
        opts = ["Off", "On"]
        band.append(
            RadioSetting(
                "txnar_vhf", "TX Narrowband VHF",
                RadioSettingValueList(opts, opts[_overlay.txnar_vhf])))
        band.append(
            RadioSetting(
                "txnar_uhf", "TX Narrowband UHF",
                RadioSettingValueList(opts, opts[_overlay.txnar_uhf])))

        # 33 Enables/disables the VFO Tracking feature
        opts = ["Off", "On"]
        common.append(
            RadioSetting(
                "vfotr", "VFO Tracking",
                RadioSettingValueList(opts, opts[_overlay.vfotr])))

        # 34 Selects the receiving mode on the VHF band
        opts = ["Inhibit (only FM)", "AM", "Auto"]
        common.append(
            RadioSetting(
                "am", "AM Mode",
                RadioSettingValueList(opts, opts[_overlay.am])))

        # Current Band
        opts = ["VHF", "UHF"]
        common.append(
            RadioSetting(
                "current_band", "Current Band",
                RadioSettingValueList(opts, opts[_overlay.current_band])))

        # Show number of VHF and UHF channels
        val = RadioSettingValueString(0, 7,
                                      str(int(self._memobj.nb_mem_used_vhf)))
        val.set_mutable(False)
        rs = RadioSetting("num_chan_vhf", "Number of VHF channels", val)
        common.append(rs)
        val = RadioSettingValueString(0, 7,
                                      str(int(self._memobj.nb_mem_used_uhf)))
        val.set_mutable(False)
        rs = RadioSetting("num_chan_uhf", "Number of UHF channels", val)
        common.append(rs)

        return setmode

    def set_settings(self, uisettings):
        _overlay = self._memobj.overlay
        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue

            try:
                name = element.get_name()
                value = element.value

                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    element.run_apply_callback()
                else:
                    setattr(_overlay, name, value)

                LOG.debug("Setting %s: %s", name, value)
            except Exception:
                LOG.debug(element.get_name())
                raise

    def _get_upper_vhf_limit(self):
        if self._memobj is None:
            # test with tox has no _memobj
            upper_vhf_limit = 120
        else:
            upper_vhf_limit = int(self._memobj.nb_mem_used_vhf)
        return upper_vhf_limit

    def _get_upper_uhf_limit(self):
        if self._memobj is None:
            # test with tox has no _memobj
            upper_uhf_limit = 120
        else:
            upper_uhf_limit = int(self._memobj.nb_mem_used_uhf)
        return upper_uhf_limit

    @classmethod
    def match_model(cls, filedata, filename):
        return filedata[0x1ec0:0x1ec0+len(cls.IDBLOCK)] == cls.IDBLOCK

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn Radio off.\n"
            "2. Connect data cable.\n"
            "3. While holding \"TONE\" and \"REV\" buttons, turn radio on.\n"
            "4. <b>After clicking OK</b>, press \"TONE\" to send image.\n")
        rp.pre_upload = _(
            "1. Turn Radio off.\n"
            "2. Connect data cable.\n"
            "3. While holding \"TONE\" and \"REV\" buttons, turn radio on.\n"
            "4. Press \"REV\" to receive image.\n"
            "5. Make sure display says \"CLONE RX\" and green led is"
            " blinking\n"
            "6. Click OK to start transfer.\n")
        return rp

    def get_sub_devices(self):
        if not self.VARIANT:
            return [FT7100RadioVHF(self._mmap),
                    FT7100RadioUHF(self._mmap)]
        else:
            return []


class FT7100RadioVHF(FT7100Radio):
    VARIANT = "VHF Band"

    def get_features(self):
        LOG.debug("VHF get_features")
        upper_vhf_limit = self._get_upper_vhf_limit()
        rf = FT7100Radio.get_features(self)
        rf.memory_bounds = (1, upper_vhf_limit)
        # Normally this band supports 120 + 10 memories. 1 based for chirpw
        rf.valid_bands = [(108000000, 180000000)]  # Supports 2-meters tx
        rf.valid_modes = MODES_VHF
        rf.valid_special_chans = ['VFO', 'Home']
        rf.has_sub_devices = False
        return rf

    def get_memory(self, number):
        LOG.debug("get_memory VHF Number: %s", number)
        if isinstance(number, int):
            if number >= 0:
                mem = FT7100Radio.get_memory(self, number + 0 - 1)
            else:
                mem = FT7100Radio.get_memory(self, number)
            mem.number = number
        else:
            mem = FT7100Radio.get_memory(self, number + '-VHF')
            mem.extd_number = number
            mem.immutable = ["number", "extd_number", "skip"]
        return mem

    def set_memory(self, mem):
        LOG.debug("set_memory VHF Number: %s", mem.number)
        # We modify memory, so dupe() it to avoid changing our caller's
        # version
        mem = mem.dupe()
        if isinstance(mem.number, int):
            if mem.number >= 0:
                mem.number += -1
        else:
            mem.number += '-VHF'
        super(FT7100RadioVHF, self).set_memory(mem)
        return


class FT7100RadioUHF(FT7100Radio):
    VARIANT = "UHF Band"

    def get_features(self):
        LOG.debug("UHF get_features")
        upper_uhf_limit = self._get_upper_uhf_limit()
        rf = FT7100Radio.get_features(self)
        rf.memory_bounds = (1, upper_uhf_limit)
        # Normally this band supports 120 + 10 memories. 1 based for chirpw
        rf.valid_bands = [(320000000, 999990000)]  # Supports 70-centimeters tx
        rf.valid_modes = MODES_UHF
        rf.valid_special_chans = ['VFO', 'Home']
        rf.has_sub_devices = False
        return rf

    def get_memory(self, number):
        LOG.debug("get_memory UHF Number: %s", number)
        upper_vhf_limit = self._get_upper_vhf_limit()
        if isinstance(number, int):
            if number >= 0:
                mem = FT7100Radio.get_memory(self, number + 10 +
                                             upper_vhf_limit - 1)
            else:
                mem = FT7100Radio.get_memory(self, number)
            mem.number = number
        else:
            mem = FT7100Radio.get_memory(self, number + '-UHF')
            mem.extd_number = number
            mem.immutable = ["number", "extd_number", "skip"]
        return mem

    def set_memory(self, mem):
        LOG.debug("set_memory UHF Number: %s", mem.number)
        # We modify memory, so dupe() it to avoid changing our caller's
        # version
        mem = mem.dupe()
        upper_vhf_limit = self._get_upper_vhf_limit()
        if isinstance(mem.number, int):
            if mem.number >= 0:
                mem.number += upper_vhf_limit - 1 + 10
        else:
            mem.number += '-UHF'
        super(FT7100RadioUHF, self).set_memory(mem)
        return



# Copyright 2011 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp.drivers import icf
from chirp import chirp_common, directory, bitwise

MEM_FORMAT = """
struct {
  u24 freq;
  ul16 offset;
  char name[6];
  u8 unknown2:2,
     rtone:6;
  u8 unknown3:2,
     ctone:6;
  u8 unknown4:1,
     dtcs:7;
  u8 tuning_step:4,
     narrow:1,
     unknown5:1,
     duplex:2;
  u8 unknown6:1,
     power:2,
     dtcs_polarity:2,
     tmode:3;
} memory[300];

#seekto 0x12E0;
u8 used[38];

#seekto 0x1306;
u8 skips[38];

#seekto 0x132C;
u8 pskips[38];

#seekto 0x1360;
struct {
  u8 bank;
  u8 index;
} banks[300];

#seekto 0x16D0;
struct {
  char name[6];
} bank_names[26];

"""

TMODES = ["", "Tone", "TSQL", "TSQL", "DTCS", "DTCS"]
DUPLEX = ["", "-", "+"]
DTCS_POLARITY = ["NN", "NR", "RN", "RR"]
TUNING_STEPS = [5.0, 5.0, 5.0, 5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0,
                50.0, 100.0, 125.0, 200.0]
POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5),
                chirp_common.PowerLevel("Low", watts=0.5),
                chirp_common.PowerLevel("Mid", watts=1.0),
                ]


class ICT70Bank(icf.IcomBank):
    """ICT70 bank"""
    def get_name(self):
        _bank = self._model._radio._memobj.bank_names[self.index]
        return str(_bank.name).rstrip()

    def set_name(self, name):
        _bank = self._model._radio._memobj.bank_names[self.index]
        _bank.name = name.ljust(6)[:6]


@directory.register
class ICT70Radio(icf.IcomCloneModeRadio):
    """Icom IC-T70"""
    VENDOR = "Icom"
    MODEL = "IC-T70"

    _model = "\x32\x53\x00\x01"
    _memsize = 0x19E0
    _endframe = "Icom Inc\x2eCF"

    _ranges = [(0x0000, 0x19E0, 32)]

    _num_banks = 26
    _bank_class = ICT70Bank

    def _get_bank(self, loc):
        _bank = self._memobj.banks[loc]
        if _bank.bank != 0xFF:
            return _bank.bank
        else:
            return None

    def _set_bank(self, loc, bank):
        _bank = self._memobj.banks[loc]
        if bank is None:
            _bank.bank = 0xFF
        else:
            _bank.bank = bank

    def _get_bank_index(self, loc):
        _bank = self._memobj.banks[loc]
        return _bank.index

    def _set_bank_index(self, loc, index):
        _bank = self._memobj.banks[loc]
        _bank.index = index

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (0, 299)
        rf.valid_tmodes = TMODES
        rf.valid_duplexes = DUPLEX
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_bands = [(136000000, 174000000), (400000000, 479000000)]
        rf.valid_skips = ["", "S", "P"]
        rf.valid_tuning_steps = TUNING_STEPS
        rf.valid_name_length = 6
        rf.has_ctone = True
        rf.has_bank = True
        rf.has_bank_index = True
        rf.has_bank_names = True
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def get_memory(self, number):
        bit = 1 << (number % 8)
        byte = int(number / 8)

        _mem = self._memobj.memory[number]
        _usd = self._memobj.used[byte]
        _skp = self._memobj.skips[byte]
        _psk = self._memobj.pskips[byte]

        mem = chirp_common.Memory()
        mem.number = number

        if _usd & bit:
            mem.empty = True
            return mem

        if _mem.freq & 0x800000:
            mem.freq = (_mem.freq & ~0x800000) * 6250
        else:
            mem.freq = _mem.freq * 5000
        mem.offset = _mem.offset * 5000
        mem.name = str(_mem.name).rstrip()
        mem.rtone = chirp_common.TONES[_mem.rtone]
        mem.ctone = chirp_common.TONES[_mem.ctone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.tuning_step = TUNING_STEPS[_mem.tuning_step]
        mem.mode = _mem.narrow and "NFM" or "FM"
        mem.duplex = DUPLEX[_mem.duplex]
        mem.power = POWER_LEVELS[_mem.power]
        mem.dtcs_polarity = DTCS_POLARITY[_mem.dtcs_polarity]
        mem.tmode = TMODES[_mem.tmode]
        mem.skip = (_psk & bit and "P") or (_skp & bit and "S") or ""

        return mem

    def set_memory(self, mem):
        bit = 1 << (mem.number % 8)
        byte = int(mem.number / 8)

        _mem = self._memobj.memory[mem.number]
        _usd = self._memobj.used[byte]
        _skp = self._memobj.skips[byte]
        _psk = self._memobj.pskips[byte]

        _mem.set_raw("\x00" * (_mem.size() // 8))

        if mem.empty:
            _usd |= bit
            return

        _usd &= ~bit

        if chirp_common.is_12_5(mem.freq):
            _mem.freq = (mem.freq // 6250) | 0x800000
        else:
            _mem.freq = mem.freq // 5000
        _mem.offset = mem.offset // 5000
        _mem.name = mem.name.ljust(6)[:6]
        _mem.rtone = chirp_common.TONES.index(mem.rtone)
        _mem.ctone = chirp_common.TONES.index(mem.ctone)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.tuning_step = TUNING_STEPS.index(mem.tuning_step)
        _mem.narrow = mem.mode == "NFM"
        _mem.duplex = DUPLEX.index(mem.duplex)
        _mem.dtcs_polarity = DTCS_POLARITY.index(mem.dtcs_polarity)
        _mem.tmode = TMODES.index(mem.tmode)
        if mem.power:
            _mem.power = POWER_LEVELS.index(mem.power)
        else:
            _mem.power = 0

        if mem.skip == "S":
            _skp |= bit
            _psk &= ~bit
        elif mem.skip == "P":
            _skp &= ~bit
            _psk |= bit
        else:
            _skp &= ~bit
            _psk &= ~bit



# Copyright 2016 Pavel Milanes CO7WT, <co7wt@frcuba.co.cu> <pavelmc@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettings

import time
import struct
import logging

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0000;
struct {
  lbcd rxfreq[4];
  lbcd txfreq[4];
} memory[32];

#seekto 0x0100;
struct {
  lbcd rx_tone[2];
  lbcd tx_tone[2];
} tone[32];

#seekto 0x0180;
struct {
  u8 unknown0:1,
     unknown1:1,
     wide:1,            // wide: 1 = wide, 0 = narrow
     power:1,           // power: 1 = high, 0 = low
     busy_lock:1,       // busy lock:  1 = off, 0 = on
     pttid:1,           // ptt id:  1 = off, 0 = on
     dtmf:1,            // dtmf signaling:  1 = off, 0 = on
     twotone:1;         // 2-tone signaling:  1 = off, 0 = on
} ch_settings[32];

#seekto 0x02B0;
struct {
    u8 unknown10[16];      // x02b0
    u8 unknown11[16];      // x02c0
    u8 active[4];            // x02d0
    u8 scan[4];              // x02d4
    u8 unknown12[8];         // x02d8
    u8 unknown13;          // x02e0
    u8 kMON;                 // 0x02d1 MON Key
    u8 kA;                   // 0x02d2 A Key
    u8 kSCN;                 // 0x02d3 SCN Key
    u8 kDA;                  // 0x02d4 D/A Key
    u8 unknown14;            // x02e5
    u8 min_vol;              // x02e6 byte 0-31 0 = off
    u8 poweron_tone;         // x02e7 power on tone 0 = off, 1 = on
    u8 tot;                  // x02e8 Time out Timer 0 = off, 1 = 30s (max 300)
    u8 unknown15[3];         // x02e9-x02eb
    u8 dealer_tuning;        // x02ec ? bit 0? 0 = off, 1 = on
    u8 clone;                // x02ed ? bit 0? 0 = off, 1 = on
    u8 unknown16[2];         // x02ee-x2ef
    u8 unknown17[16];      // x02f0
    u8 unknown18[5];       // x0300
    u8 clear2transpond;      // x0305 byte 0 = off, 1 = on
    u8 off_hook_decode;      // x0306 byte 0 = off, 1 = on
    u8 off_hook_hornalert;   // x0307 byte 0 = off, 1 = on
    u8 unknown19[8];         // x0308-x030f
    u8 unknown20[16];      // x0310
} settings;
"""

KEYS = {
    0x00: "Disabled",
    0x01: "Monitor",
    0x02: "Talk Around",
    0x03: "Horn Alert",
    0x04: "Public Address",
    0x05: "Auxiliary",
    0x06: "Scan",
    0x07: "Scan Del/Add",
    0x08: "Home Channel",
    0x09: "Operator Selectable Tone",
    0x0C: "Unknown"
}

MEM_SIZE = 0x400
BLOCK_SIZE = 8
MEM_BLOCKS = list(range(0, (MEM_SIZE // BLOCK_SIZE)))
ACK_CMD = b"\x06"
TIMEOUT = 1

POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=1),
                chirp_common.PowerLevel("High", watts=5)]
MODES = ["NFM", "FM"]
SKIP_VALUES = ["", "S"]
TONES = list(chirp_common.TONES)
TONES.remove(254.1)
DTCS_CODES = chirp_common.DTCS_CODES

TOT = ["off"] + ["%s" % x for x in range(30, 330, 30)]
VOL = ["off"] + ["%s" % x for x in range(1, 32)]

debug = False


def rawrecv(radio, amount):
    """Raw read from the radio device"""
    data = b""
    try:
        data = radio.pipe.read(amount)
        if debug:
            print("<= %02i: %s" % (len(data), util.hexprint(data)))
    except:
        raise errors.RadioError("Error reading data from radio")

    return data


def rawsend(radio, data):
    """Raw send to the radio device"""
    try:
        radio.pipe.write(data)
        if debug:
            print("=> %02i: %s" % (len(data), util.hexprint(data)))
    except:
        raise errors.RadioError("Error sending data from radio")


def send(radio, frame):
    """Generic send data to the radio"""
    rawsend(radio, frame)


def make_frame(cmd, addr, data=b""):
    """Pack the info in the format it likes"""
    ts = struct.pack(">BHB", ord(cmd), addr, 8)
    if data == b"":
        return ts
    else:
        if len(data) == 8:
            return ts + data
        else:
            raise errors.InvalidValueError("Data length of unexpected length")


def handshake(radio, msg="", full=False):
    """Make a full handshake, if not full just hals"""
    # send ACK if commandes
    if full is True:
        rawsend(radio, ACK_CMD)
    # receive ACK
    ack = rawrecv(radio, 1)
    # check ACK
    if ack != ACK_CMD:
        # close_radio(radio)
        mesg = "Handshake failed: " + msg
        raise errors.RadioError(mesg)


def recv(radio):
    """Receive data from the radio, 12 bytes, 4 in the header, 8 as data"""
    rxdata = rawrecv(radio, 12)

    if len(rxdata) != 12:
        raise errors.RadioError(
            "Received a length of data that is not possible")

    cmd, addr, length = struct.unpack(">BHB", rxdata[0:4])
    data = ""
    if length == 8:
        data = rxdata[4:]

    return data


def open_radio(radio):
    """Open the radio into program mode and check if it's the correct model"""
    # Set serial discipline
    try:
        radio.pipe.baudrate = 9600
        radio.pipe.parity = "N"
        radio.pipe.timeout = 0.5
        LOG.debug("Serial port open successful")
    except:
        msg = "Serial error: Can't set serial line discipline"
        raise errors.RadioError(msg)

    magic = b"PROGRAM"
    LOG.debug("Sending MAGIC")
    exito = False

    # The radio sends a continuous stream of 0xFF while idling in normal
    # operating mode. Chew those up so we can pace it below.
    devnull = rawrecv(radio, 256)

    # "Pace" the radio's output by sending one byte of the magic string for
    # every 0xFF we get from the radio
    for b in magic:
        rawrecv(radio, 1)
        rawsend(radio, bytes([b]))

    # Read more than we need here because we may have received additional bytes
    # in between the flush and the start of the magic. Consider it successful
    # if we see ACK anywhere in the response
    resp = rawrecv(radio, 256)
    if ACK_CMD not in resp:
        raise errors.RadioError('Radio did not respond to program mode')

    # 0x02 Means "show me your ident"
    rawsend(radio, b"\x02")
    ident = rawrecv(radio, 8)

    # validate the input
    if len(ident) != 8:
        LOG.debug("Wrong ID, get only %s bytes, we expect 8" % len(ident))
        LOG.debug(util.hexprint(ident))
        msg = "Bad ID received, just %s bytes, we want 8" % len(ident)
        raise errors.RadioError(msg)

    handshake(radio, "Comm error after ident", True)
    LOG.debug("Correct get ident and handshake")

    if not (radio.TYPE in ident):
        LOG.debug("Incorrect model ID:")
        LOG.debug(util.hexprint(ident))
        msg = "Incorrect model ID, got %s, it not contains %s" % \
            (ident[0:5], radio.TYPE)
        raise errors.RadioError(msg)

    LOG.debug("Full ident string is:")
    LOG.debug(util.hexprint(ident))


def do_download(radio):
    """This is your download function"""
    open_radio(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE // BLOCK_SIZE
    status.msg = "Cloning from radio..."
    radio.status_fn(status)

    data = b""
    LOG.debug("Starting the download")
    for addr in MEM_BLOCKS:
        send(radio, make_frame(b"R", addr * BLOCK_SIZE))
        data += recv(radio)
        handshake(radio, "Rx error in block %03i" % addr, True)
        LOG.debug("Block: %04x, Pos: %06x" % (addr, addr * BLOCK_SIZE))

        # UI Update
        status.cur = addr
        status.msg = "Cloning from radio..."
        radio.status_fn(status)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    """Upload info to radio"""
    open_radio(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE // BLOCK_SIZE
    status.msg = "Cloning to radio..."
    radio.status_fn(status)
    count = 0

    for addr in MEM_BLOCKS:
        # UI Update
        status.cur = addr
        status.msg = "Cloning to radio..."
        radio.status_fn(status)

        pos = addr * BLOCK_SIZE
        if pos > 0x0378:
            # it seems that from this point forward is read only !?!?!?
            continue

        data = radio.get_mmap()[pos:pos + BLOCK_SIZE]
        send(radio, make_frame(b"W", pos, data))
        LOG.debug("Block: %04x, Pos: %06x" % (addr, pos))

        time.sleep(0.1)
        handshake(radio, "Rx error in block %04x" % addr)


def get_rid(data):
    """Extract the radio identification from the firmware"""
    rid = data[0x0378:0x0380]
    # we have to invert rid
    nrid = rid[::-1]

    return nrid


def model_match(cls, data):
    """Match the opened/downloaded image to the correct version"""
    rid = get_rid(data)

    if debug:
        print("Full ident string is %s" % util.hexprint(rid))

    if (rid in cls.VARIANTS):
        # correct model
        return True
    else:
        return False


class Kenwood_M60_Radio(chirp_common.CloneModeRadio,
                        chirp_common.ExperimentalRadio):
    """Kenwood Mobile Family 60 Radios"""
    VENDOR = "Kenwood"
    _range = [136000000, 500000000]  # don't mind, it will be overwritten
    _upper = 32
    VARIANT = ""
    MODEL = ""
    NEEDS_COMPAT_SERIAL = False

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            'This driver is experimental: Use at your own risk!'
        rp.pre_download = _(
            "Follow these instructions to read your radio:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Click OK to start\n")
        rp.pre_upload = _(
            "Follow these instructions to write your radio:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Click OK to start\n")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.has_name = False
        rf.has_offset = True
        rf.has_mode = True
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_modes = MODES
        rf.valid_duplexes = ["", "-", "+", "off"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_skips = SKIP_VALUES
        rf.valid_dtcs_codes = DTCS_CODES
        rf.valid_bands = [self._range]
        rf.memory_bounds = (1, self._upper)
        rf.valid_tuning_steps = [5., 6.25, 10., 12.5]
        return rf

    def sync_in(self):
        """Download from radio"""
        try:
            self._mmap = do_download(self)
        except Exception as e:
            LOG.exception('Failed to download: %s', e)
            rawsend(self, b'E')
            raise errors.RadioError('Error downloading from radio')
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        # Get the data ready for upload
        try:
            self._prep_data()
        except Exception as e:
            LOG.exception('Failed to prep data: %s', e)
            raise errors.RadioError("Error processing the radio data")

        # do the upload
        try:
            do_upload(self)
        except Exception as e:
            LOG.exception('Failed to upload: %s', e)
            rawsend(self, b'E')
            raise errors.RadioError("Error uploading data to radio")

    def set_variant(self):
        """Select and set the correct variables for the class according
        to the correct variant of the radio"""
        rid = get_rid(self._mmap)

        # identify the radio variant and set the environment to its values
        try:
            self._upper, low, high, self._kind = self.VARIANTS[rid]

            # Frequency ranges: some model/variants are able to work the near
            # ham bands, even if they are outside the OEM ranges.
            # By experimentation we found that a +/- 4% at the edges is in most
            # cases safe and will cover the near ham band in full
            self._range = [low * 1000000 * 0.96, high * 1000000 * 1.04]

            # put the VARIANT in the class, clean the model / CHs / Type
            # in the same layout as the KPG program
            self._VARIANT = self.MODEL + " [" + str(self._upper) + "CH]: "
            # In the OEM string we show the real OEM ranges
            self._VARIANT += self._kind + ", %d - %d MHz" % (low, high)

        except KeyError:
            LOG.debug("Wrong Kenwood radio, ID or unknown variant")
            LOG.debug(util.hexprint(rid))
            raise errors.RadioError(
                "Wrong Kenwood radio, ID or unknown variant, see LOG output.")

    def _prep_data(self):
        """Prepare the areas in the memmap to do a consistent write
        it has to make an update on the x200 flag data"""
        achs = 0

        for i in range(0, self._upper):
            if self.get_active(i) is True:
                achs += 1

        # The x0200 area has the settings for the DTMF/2-Tone per channel,
        # as by default any of this radios has the DTMF IC installed;
        # we clean this areas
        fldata = b"\x00\xf0\xff\xff\xff" * achs + \
            b"\xff" * (5 * (self._upper - achs))
        self._fill(0x0200, fldata)

    def _fill(self, offset, data):
        """Fill an specified area of the memmap with the passed data"""
        self._mmap[offset] = data

    def process_mmap(self):
        """Process the mem map into the mem object"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)
        # to set the vars on the class to the correct ones
        self.set_variant()

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def decode_tone(self, val):
        """Parse the tone data to decode from mem, it returns:
        Mode (''|DTCS|Tone), Value (None|###), Polarity (None,N,R)"""
        if val.get_raw(asbytes=False) == "\xFF\xFF":
            return '', None, None

        val = int(val)
        if val >= 12000:
            a = val - 12000
            return 'DTCS', a, 'R'
        elif val >= 8000:
            a = val - 8000
            return 'DTCS', a, 'N'
        else:
            a = val / 10.0
            return 'Tone', a, None

    def encode_tone(self, memval, mode, value, pol):
        """Parse the tone data to encode from UI to mem"""
        if mode == '':
            memval[0].set_raw(0xFF)
            memval[1].set_raw(0xFF)
        elif mode == 'Tone':
            memval.set_value(int(value * 10))
        elif mode == 'DTCS':
            flag = 0x80 if pol == 'N' else 0xC0
            memval.set_value(value)
            memval[1].set_bits(flag)
        else:
            raise Exception("Internal error: invalid mode `%s'" % mode)

    def get_scan(self, chan):
        """Get the channel scan status from the 4 bytes array on the eeprom
        then from the bits on the byte, return '' or 'S' as needed"""
        result = "S"
        byte = int(chan/8)
        bit = chan % 8
        res = self._memobj.settings.scan[byte] & (pow(2, bit))
        if res > 0:
            result = ""

        return result

    def set_scan(self, chan, value):
        """Set the channel scan status from UI to the mem_map"""
        byte = int(chan/8)
        bit = chan % 8

        # get the actual value to see if I need to change anything
        actual = self.get_scan(chan)
        if actual != value:
            # I have to flip the value
            rbyte = self._memobj.settings.scan[byte]
            rbyte = rbyte ^ pow(2, bit)
            self._memobj.settings.scan[byte] = rbyte

    def get_active(self, chan):
        """Get the channel active status from the 4 bytes array on the eeprom
        then from the bits on the byte, return True/False"""
        byte = int(chan/8)
        bit = chan % 8
        res = self._memobj.settings.active[byte] & (pow(2, bit))
        return bool(res)

    def set_active(self, chan, value=True):
        """Set the channel active status from UI to the mem_map"""
        byte = int(chan/8)
        bit = chan % 8

        # get the actual value to see if I need to change anything
        actual = self.get_active(chan)
        if actual != bool(value):
            # I have to flip the value
            rbyte = self._memobj.settings.active[byte]
            rbyte = rbyte ^ pow(2, bit)
            self._memobj.settings.active[byte] = rbyte

    def get_memory(self, number):
        """Get the mem representation from the radio image"""
        _mem = self._memobj.memory[number - 1]
        _tone = self._memobj.tone[number - 1]
        _ch = self._memobj.ch_settings[number - 1]

        # Create a high-level memory object to return to the UI
        mem = chirp_common.Memory()

        # Memory number
        mem.number = number

        if (_mem.get_raw(asbytes=False)[0] == "\xFF" or
                not self.get_active(number - 1)):
            mem.empty = True
            # but is not enough, you have to clear the memory in the mmap
            # to get it ready for the sync_out process
            _mem.set_raw("\xFF" * 8)
            return mem

        # Freq and offset
        mem.freq = int(_mem.rxfreq) * 10
        # tx freq can be blank
        if _mem.get_raw(asbytes=False)[4] == "\xFF":
            # TX freq not set
            mem.offset = 0
            mem.duplex = "off"
        else:
            # TX feq set
            offset = (int(_mem.txfreq) * 10) - mem.freq
            if offset < 0:
                mem.offset = abs(offset)
                mem.duplex = "-"
            elif offset > 0:
                mem.offset = offset
                mem.duplex = "+"
            else:
                mem.offset = 0

        # power
        mem.power = POWER_LEVELS[_ch.power]

        # wide/marrow
        mem.mode = MODES[_ch.wide]

        # skip
        mem.skip = self.get_scan(number - 1)

        # tone data
        rxtone = txtone = None
        txtone = self.decode_tone(_tone.tx_tone)
        rxtone = self.decode_tone(_tone.rx_tone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        # Extra
        # bank and number in the channel
        mem.extra = RadioSettingGroup("extra", "Extra")

        bl = RadioSetting("busy_lock", "Busy Channel lock",
                          RadioSettingValueBoolean(
                              not bool(_ch.busy_lock)))
        mem.extra.append(bl)

        return mem

    def set_memory(self, mem):
        """Set the memory data in the eeprom img from the UI
        not ready yet, so it will return as is"""

        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[mem.number - 1]
        _tone = self._memobj.tone[mem.number - 1]
        _ch = self._memobj.ch_settings[mem.number - 1]

        # Empty memory
        if mem.empty:
            _mem.set_raw("\xFF" * 8)
            # empty the active bit
            self.set_active(mem.number - 1, False)
            return

        # freq rx
        _mem.rxfreq = mem.freq / 10

        # freq tx
        if mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        elif mem.duplex == "off":
            for byte in _mem.txfreq:
                byte.set_raw("\xFF")
        else:
            _mem.txfreq = mem.freq / 10

        # tone data
        ((txmode, txtone, txpol), (rxmode, rxtone, rxpol)) = \
            chirp_common.split_tone_encode(mem)
        self.encode_tone(_tone.tx_tone, txmode, txtone, txpol)
        self.encode_tone(_tone.rx_tone, rxmode, rxtone, rxpol)

        # power, default power is low
        if mem.power is None:
            mem.power = POWER_LEVELS[0]

        _ch.power = POWER_LEVELS.index(mem.power)

        # wide/marrow
        _ch.wide = MODES.index(mem.mode)

        # skip
        self.set_scan(mem.number - 1, mem.skip)

        # extra settings
        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

        # set the mem a active in the _memmap
        self.set_active(mem.number - 1)

        return mem

    @classmethod
    def match_model(cls, filedata, filename):
        match_size = False
        match_model = False

        # testing the file data size
        if len(filedata) == MEM_SIZE:
            match_size = True

        # testing the firmware model fingerprint
        match_model = model_match(cls, filedata)

        if match_size and match_model:
            return True
        else:
            return False

    def get_settings(self):
        """Translate the bit in the mem_struct into settings in the UI"""
        sett = self._memobj.settings

        # basic features of the radio
        basic = RadioSettingGroup("basic", "Basic Settings")
        # buttons
        fkeys = RadioSettingGroup("keys", "Front keys config")

        top = RadioSettings(basic, fkeys)

        # Basic
        val = RadioSettingValueString(0, 35, self._VARIANT)
        val.set_mutable(False)
        mod = RadioSetting("not.mod", "Radio version", val)
        basic.append(mod)

        tot = RadioSetting("settings.tot", "Time Out Timer (TOT)",
                           RadioSettingValueList(TOT, sett.tot))
        basic.append(tot)

        minvol = RadioSetting("settings.min_vol", "Minimum volume",
                              RadioSettingValueList(VOL,
                                                    sett.min_vol))
        basic.append(minvol)

        ptone = RadioSetting("settings.poweron_tone", "Power On tone",
                             RadioSettingValueBoolean(
                                 bool(sett.poweron_tone)))
        basic.append(ptone)

        sprog = RadioSetting("settings.dealer_tuning", "Dealer Tuning",
                             RadioSettingValueBoolean(
                                 bool(sett.dealer_tuning)))
        basic.append(sprog)

        clone = RadioSetting("settings.clone", "Allow clone",
                             RadioSettingValueBoolean(
                                 bool(sett.clone)))
        basic.append(clone)

        # front keys
        rs = RadioSettingValueList(KEYS.values(),
                                   KEYS[int(sett.kMON)])
        mon = RadioSetting("settings.kMON", "MON", rs)
        fkeys.append(mon)

        rs = RadioSettingValueList(KEYS.values(),
                                   KEYS[int(sett.kA)])
        a = RadioSetting("settings.kA", "A", rs)
        fkeys.append(a)

        rs = RadioSettingValueList(KEYS.values(),
                                   KEYS[int(sett.kSCN)])
        scn = RadioSetting("settings.kSCN", "SCN", rs)
        fkeys.append(scn)

        rs = RadioSettingValueList(KEYS.values(),
                                   KEYS[int(sett.kDA)])
        da = RadioSetting("settings.kDA", "D/A", rs)
        fkeys.append(da)

        return top

    def set_settings(self, settings):
        """Translate the settings in the UI into bit in the mem_struct
        I don't understand well the method used in many drivers
        so, I used mine, ugly but works ok"""

        mobj = self._memobj

        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue

            # Let's roll the ball
            if "." in element.get_name():
                inter, setting = element.get_name().split(".")
                # you must ignore the settings with "not"
                # this are READ ONLY attributes
                if inter == "not":
                    continue

                obj = getattr(mobj, inter)
                value = element.value

                # case keys, with special config
                if setting[0] == "k":
                    value = list(KEYS.keys())[
                            list(KEYS.values()).index(str(value))]

                # integers case + special case
                if setting in ["tot", "min_vol"]:
                    # catching the "off" values as zero
                    try:
                        value = int(value)
                    except:
                        value = 0

                # Bool types + inverted
                if setting in ["poweron_tone", "dealer_tuning", "clone"]:
                    value = bool(value)

            # Apply al configs done
            # DEBUG
            # print("%s: %s" % (setting, value))
            setattr(obj, setting, value)


# This are the oldest family 60 models (Black keys), just mobiles support here

@directory.register
class TK760_Radio(Kenwood_M60_Radio):
    """Kenwood TK-760 Radios"""
    MODEL = "TK-760"
    TYPE = b"M0760"
    VARIANTS = {
        b"M0760\x01\x00\x00": (32, 136, 156, "K2"),
        b"M0760\x00\x00\x00": (32, 148, 174, "K")
        }


@directory.register
class TK762_Radio(Kenwood_M60_Radio):
    """Kenwood TK-762 Radios"""
    MODEL = "TK-762"
    TYPE = b"M0762"
    VARIANTS = {
        b"M0762\x01\x00\x00": (2, 136, 156, "K2"),
        b"M0762\x00\x00\x00": (2, 148, 174, "K")
        }


@directory.register
class TK768_Radio(Kenwood_M60_Radio):
    """Kenwood TK-768 Radios"""
    MODEL = "TK-768"
    TYPE = b"M0768"
    VARIANTS = {
        b"M0768\x21\x00\x00": (32, 136, 156, "K2"),
        b"M0768\x20\x00\x00": (32, 148, 174, "K")
        }


@directory.register
class TK860_Radio(Kenwood_M60_Radio):
    """Kenwood TK-860 Radios"""
    MODEL = "TK-860"
    TYPE = b"M0860"
    VARIANTS = {
        b"M0860\x05\x00\x00": (32, 406, 430, "F4"),
        b"M0860\x04\x00\x00": (32, 488, 512, "F3"),
        b"M0860\x03\x00\x00": (32, 470, 496, "F2"),
        b"M0860\x02\x00\x00": (32, 450, 476, "F1")
        }


@directory.register
class TK862_Radio(Kenwood_M60_Radio):
    """Kenwood TK-862 Radios"""
    MODEL = "TK-862"
    TYPE = b"M0862"
    VARIANTS = {
        b"M0862\x05\x00\x00": (2, 406, 430, "F4"),
        b"M0862\x04\x00\x00": (2, 488, 512, "F3"),
        b"M0862\x03\x00\x00": (2, 470, 496, "F2"),
        b"M0862\x02\x00\x00": (2, 450, 476, "F1")
        }


@directory.register
class TK868_Radio(Kenwood_M60_Radio):
    """Kenwood TK-868 Radios"""
    MODEL = "TK-868"
    TYPE = b"M0868"
    VARIANTS = {
        b"M0868\x25\x00\x00": (32, 406, 430, "F4"),
        b"M0868\x24\x00\x00": (32, 488, 512, "F3"),
        b"M0868\x23\x00\x00": (32, 470, 496, "F2"),
        b"M0868\x22\x00\x00": (32, 450, 476, "F1")
        }
